//  Copyright (C) 2005 Alexander S. Ovcharenko, SymbioForge.
//
//  FlushFdb - utility which allows to extract raster PalmOS fonts (type 1
//  and type 2) from .pdb database to a binary raw resource files.
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
//  gcc flushfdb.c toollib.c -o flushfdb
//
//   Where GCC is a GNU C compiler or other used by your computer operation
//  system;

#include <stdio.h>
#include "toollib.h"

static		void usage( void );
pdbHeaderType	dbHeader;

main( int argc, char *argv[] )
{ int		returnStatus = 0;
  int   	i, resCounter, arglen;
  int		noFileName = 1;
  Boolean	customResourceNumber=false, checkFontSign=true, debugExtract=false;
  unsigned int  resBeginFrom = 1000;

  puts("FlushFdb - Exports font data of Palm OS(tm) database to font binary resources.");
  puts("Version 1.3 (c) 2005, SymbioForge (http://www.symbioforge.com/feii/).");

  for ( i=1; i < argc; i++ )
   { 
    arglen = strlen( argv[i] ); // Get length of this argument;

    // Treat command line options;
    if ( (arglen >= 1 ) && (argv[i][0]=='-' || argv[i][0]=='/') ) // Assume this is a flag;
       {  unsigned int value=0;
          if ( arglen >= 2 ) // Check the flag;
	    {
	     switch( argv[i][1] )
               {
		case 'h':
		case 'H':
		case '?':
			usage();
		break;

		case 'f':
		case 'F':
			checkFontSign=false;			
		break;

		case 'D':
		case 'd':
			debugExtract=true;
		break;	

		case 'N':
                case 'n':
			if ( arglen >=3 ) // Get value;
			 {
			  value=atoi( &argv[i][2] );
/*
			  if ( value < 1000 )
				{
                             	 puts( "\nError: resource value is less than 1000. Program stopped.\00" );
				 errorExit( errorExitValue );
				}
			  else
*/
				{
				 customResourceNumber=true;
				 resBeginFrom = value;
				};
		         }
                        else
			 {
                          puts( "\nError: resource start value is not set. Program stopped.\00" );
			  errorExit( errorExitValue );
			 };
                break;

                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
               };
          };
      }
     else
      { FILE		* inFilePtr, * outFilePtr;
	char		* resFileNamePtr=NULL;
	unsigned long	listEntryOffset=0l, pdbRecordOffset=0l, nextPdbRecordOffset=0l;
	unsigned long	fileSize, recordSize;
	UInt16		originalResourceID=1000;
	UInt32		originalResourceType;
	unsigned int	recordIndex;
	int		databaseFormat= unknownDataFormat;
	FontTypeV2	fontHeaderV2;

	noFileName = 0; // So we have a file name used in argument.

	inFilePtr = openFile( argv[i], "rb" );
	if ( inFilePtr == NULL ) 
		{ 
		  printf("\nError: can not open file '%s'. Program aborted.\n\00", argv[i] );
		  errorExit( errorFileOperation );
                };

	fileSize = fileLength( inFilePtr ); // Get size of opened file;
	if ( fileSize < pdbHeaderLength )
		{ 
		  printf("\nError: size of file is less than length of .pdb header. Program aborted.\n\00");
		  errorExit( errorFileOperation );
		};

	fileGetPdbHeader( &dbHeader, inFilePtr );
	if ( (UInt32)dbHeader.dbType!=(UInt32)fontDatabaseID0 &&
	     (UInt32)dbHeader.dbType!=(UInt32)fontDatabaseID1 &&
             (UInt32)dbHeader.dbType!=(UInt32)fontDatabaseID2 &&
		checkFontSign )
	   {
	    	printf("\nFile '%s' is not a Palm OS font database. Program stopped.\n", argv[i] );
	    	errorExit( errorExitValue );
	   };

	if ( fileGetPDBNumRecords( 0, inFilePtr ) == 0 )
	   {
	    	printf("\nError: there is no data in '%s' database. Program stopped.\n", argv[i] );
	    	errorExit( errorExitValue );
	   };

	// Get first listEntry;
	(unsigned long)listEntryOffset = (unsigned long)fileGetPDBListEntryOffset( 0, inFilePtr );
	// Get format of database ( records or resources );
	databaseFormat = fileGetPDBResourceFormat( listEntryOffset, inFilePtr );

	// Check for mutiple record lists. It is unusal if it have offset on the next recordlist.
	if ( ((databaseFormat!=resourcesDataFormat) && (databaseFormat!=recordsDataFormat)) || ( fileGetPDBNextRecordListOffset( (unsigned long)listEntryOffset, inFilePtr )!=0l ) )
	     {
		// printf( "Data format is %u\n", databaseFormat);
	    	printf("\nError: unsupported format of '%s' font database. Program stopped.\n", argv[i] );
	    	errorExit( errorExitValue );
	     };

	printf("\nExport from: '%s' (%s database).\n", &dbHeader.dbName, databaseFormat==resourcesDataFormat? "resources": "records");
	recordIndex = 0; // Initiate record treat from 0;
	resFileNamePtr=(char *)malloc( dbNameLength+1 ); // Allocate memory for resource name.

	do // Strip a database record by record, resource by resource;
	 {
	   // Get a record offset.
	   if ( databaseFormat==recordsDataFormat )
	   	(UInt32)pdbRecordOffset = (UInt32)fileGetPDBRecordOffset( (UInt32)listEntryOffset, recordIndex++, inFilePtr );
	   else
  	   	(UInt32)pdbRecordOffset = fileGetResourceOffset( (UInt32)listEntryOffset, recordIndex++, inFilePtr, &originalResourceType, &originalResourceID );

	   if ( (UInt32)pdbRecordOffset != 0 ) 
		{ UInt16	fontTypeValue;
		  Boolean	isAFNX=false, isFontTypeV1, isFontTypeV2;

		  (UInt32)nextPdbRecordOffset = (databaseFormat==recordsDataFormat) ? fileGetPDBRecordOffset( (UInt32)listEntryOffset, recordIndex, inFilePtr ) : fileGetResourceOffset( (UInt32)listEntryOffset, recordIndex, inFilePtr, NULL, NULL );
		  (UInt32)recordSize = (nextPdbRecordOffset == 0) ?  (UInt32)fileSize - (UInt32)pdbRecordOffset : (UInt32)nextPdbRecordOffset-(UInt32)pdbRecordOffset;

		  printf( "#%05u", recordIndex ); // Display record number.
		  // Seek at record position and get a font type.
		  // printf( "Record offset: %lu\n", (unsigned long)pdbRecordOffset );
		  fileSeek( inFilePtr, (unsigned long)pdbRecordOffset );
		  fontTypeValue = fileGetWord( inFilePtr );

		  // Exam for afnx font record (Is it intel byte sequence format).
		  isAFNX = ((UInt16)fontTypeValue == 0x0090) || ((UInt16)fontTypeValue == 0x0092);
		  // Treat font records only.
		  isFontTypeV1 = isAFNX ? ((UInt16)fontTypeValue == 0x0090) : ((UInt16)fontTypeValue == 0x9000);
		  isFontTypeV2 = isAFNX ? ((UInt16)fontTypeValue == 0x0092) : ((UInt16)fontTypeValue == 0x9200);


		  if ( isFontTypeV1 || isFontTypeV2 || debugExtract)
			{
		  		resFileNamePtr[ 0 ] = 0; // Make an empty string.
				if ( isAFNX )
					{ UInt16 resourceID;

					 if ( databaseFormat==recordsDataFormat )
					     resourceID = resBeginFrom;
					 else
					     resourceID = customResourceNumber ? resBeginFrom: originalResourceID;
		  			 sprintf( resFileNamePtr, "%s%04x.bin", isFontTypeV1 ? "AFNX" : "afnx", resourceID );
					 printf(", font: %s (%s) ID:%u, from offset %lu size %lu bytes.", resFileNamePtr, isFontTypeV1 ? "v1" : "v2", resourceID, (UInt32)pdbRecordOffset, (UInt32)recordSize );
			    		}
				else
					{
		  			 sprintf( resFileNamePtr, "%s%04x.bin", isFontTypeV1 ? "NFNT" : "nfnt", resBeginFrom );
					 printf(", font: %s (%s) ID:%u, from offset %lu size %lu bytes.", resFileNamePtr, isFontTypeV1 ? "v1" : "v2", resBeginFrom, (UInt32)pdbRecordOffset, (UInt32)recordSize );
					};

				outFilePtr = openFile( resFileNamePtr, "wb" );
				proceedBinaryDataFileTransfer( inFilePtr, outFilePtr, (unsigned long)pdbRecordOffset, (unsigned long)recordSize );
				fclose( outFilePtr );
				puts("");

				// Increase number(ID) of resource when it is stored as a record.
				if (databaseFormat==recordsDataFormat || customResourceNumber )
				   resBeginFrom++;
			}
		  else
			 {
				// if (debugExtract) printf(" (%04x)", (UInt16)fontTypeValue );
				printf(" - is not a font %s, skiped.\n", (databaseFormat==recordsDataFormat) ? "record" : "resource" );
			 };
		};
	 } while( pdbRecordOffset!=0 );

	fclose( inFilePtr ); // Close input file;
	free( resFileNamePtr );

	customResourceNumber=false;
      };
  };
if ( argc==1 ) usage(); // Display help list;

 return returnStatus; 
};

// Display user's help.
static void usage( void )
{
  // puts("* Extracts NFNT font records to a nfntxxxx.bin files, where xxxx a hexadecimal\nnumber of resurce begining from 0x3E8 (decimal 1000).\n");
  puts("\nUsage ([] means optional):");
  puts("   flushfdb [flags] <file.pdb>");
  puts("               ... where file.[pdb|prc] is a Palm OS font databases.\n");
  puts("Flags are:\n");
  puts("-n<value>  a base <value> of resource number assigned for output files.");
  puts("-f         force to extract font resources even it is not a font database.");
  puts("-h, -?	   help screen.");
};
