//  Copyright (C) 2005 Alexander S. Ovcharenko, SymbioForge
//
//  flushfv2 - Allows to separate density parts of font resource version 2
//  to font resources version 1.
//
//  This program is free software; you can redistribute it and/or modify 
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
//  gcc flushfv2.c toollib.c -o flushfv2
//
//   Where GCC is a GNU C compiler or other used by your computer operation
//  system;

#include <stdio.h>
#include "toollib.h"

static	void usage( void );

main( int argc, char *argv[] )
{
  short		i, arglen;
  UInt16	baseResource=1000;
  UInt16	desiredFontHeaderLength=0;
  UInt16	desiredDensityEntryLength=0;
  Boolean	customBaseResource=false;
  Boolean	returnStatus=0;
  Boolean	headerFormatV5=false;

  puts("flushfv2 - Separates font v2 density parts to binary nfnt font resources v1.");
  puts("Version 1.3 (c) 2005, SymbioForge (http://www.symbioforge.com/feii/).");

  // Collect command line arguments.
  for ( i=1; i < argc; i++ )
   { 
    arglen = strlen( argv[ i ] ); // Get length of this argument;

    // Treat command line options;
    if ( (arglen >= 1 ) && (argv[i][0]=='-' || argv[i][0]=='/') ) // Assume this is a flag;
       {  unsigned int value=0;
          if ( arglen >= 2 ) // Check the flag;
	    {
	     switch( argv[i][1] )
               {
		case 'h':
		case 'H':
		case '?':
			usage();
		break;

		case 'N': // Base resource number.
                case 'n':
			if (arglen <=2) 
			   {
                   		puts( "\nError: base resource number is not specified. Program stopped.\00" );
				errorExit( errorExitValue );
			   }
			else
			  {
				baseResource=atoi( &argv[i][2] );
				customBaseResource=true;
			  };
                break;
                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
               };
          }
	  else
	  {
		puts( "\nError: incorrect use of argument. Program stopped.\00" );
		errorExit( errorExitValue );
	  };
      }
     else
      { FILE		* inFilePtr, * tmpFilePtr, * outFilePtr;
	char		outFileNameStr[ dosFileNameLength ];
	char 		* outFileNamePtr=NULL; // File name of output file.
	unsigned long	fileSize, dataToWriteSize=0;
	UInt8		bitsInRow;
	UInt16		resourceID=1000, resizeByPercent, fontResType, tablesLength=0, i0, i1, wordValue;
	UInt16		prevBitmapLocation, bitmapLocation, newBitmapLocation, glyphWidth, glyphKern;
	UInt32		totalGlyphWidth=0, imagePartLength;
	UInt32		fntBitmapLocationV1Pos=0, fntBitmapLocationV2Pos=0;

	UInt16		originalBitmapLocationTable[ maxTableElements ], correctedBitmapLocationTable[ maxTableElements ];
	UInt16		originalOffsetWidthTable[ maxTableElements ], correctedOffsetWidthTable[ maxTableElements ];

	UInt8		originalGlyphData[ bufferLength ];
	glyphBufferType	* glyphDataPtr = NULL;

	FontType	fontHeaderV1;
	FontTypeV2	fontHeaderV2;
	FontDensityType fontDensity;

	glyphDataPtr = (glyphBufferType *)&originalGlyphData[ 0 ];


	inFilePtr = openFile( argv[i], "rb" );
	if ( inFilePtr==NULL )
	   { 
		printf("\nError: can not open file '%s'. Program aborted.\n\00", argv[i] );
		errorExit( errorFileOperation );
           };

	fontResType=getFontResourceType( inFilePtr, true );
	if ( fontResType != type2nfnt )
	   {
		printf("\nError: file '%s' is not a nfnt font resource. Program stopped.\n\00", argv[i] );
		errorExit( errorFileOperation );
	   };

	printf("\nSeparate font resource: '%s'", argv[i] );
	getFontHeaderV1( inFilePtr, (FontType *)&fontHeaderV2 );
	fontHeaderV2.version = fileGetWord( inFilePtr );
	fontHeaderV2.densityCount = fileGetWord( inFilePtr );

	if ( fontHeaderV2.densityCount>0xFF || fontHeaderV2.densityCount==0 )
	   {
		printf("\nError: invalid number of densities in '%s' font resource. Program stopped.\n\00", argv[i] );
		errorExit( errorFileOperation );
	   };

	tablesLength=( fontHeaderV2.lastGlyph-fontHeaderV2.firstGlyph )+3;
	if ( customBaseResource ) 
	 {
	   resourceID=baseResource;
	   customBaseResource=false;
	 };

	// Determine font header format in the file: PalmOS(tm) 5.x or Cobalt(tm)
	if ( fontHeaderV2.owTLoc==(fontTag2V5Length+fontHeaderV2.densityCount*fontDensityV5Length+tablesLength*2)/2-8 )
	      {
		headerFormatV5=true;
		desiredFontHeaderLength=fontTag2V5Length;
		desiredDensityEntryLength = fontDensityV5Length;
		printf(" PalmOS(tm) v5.x font format.");
	      }
        else
	if ( fontHeaderV2.owTLoc==(fontTag2V6Length+fontHeaderV2.densityCount*fontDensityV6Length+tablesLength*2)/2-8 )

	      {
		headerFormatV5=false;
		desiredFontHeaderLength=fontTag2V6Length;
		desiredDensityEntryLength = fontDensityV6Length;
		fontHeaderV2.reserved = fileGetWord( inFilePtr );
		printf(" Cobalt(tm) font format.");
	      }
	else
	      {
		printf("\nError: invalid font header in '%s' font resource. Program stopped.\n\00", argv[i] );
		errorExit( errorFileOperation );
	      };

	// Seek to the font tables position. 	// Read font tables.
	fileSeek( inFilePtr, (UInt32)desiredFontHeaderLength + (fontHeaderV2.densityCount * desiredDensityEntryLength) );
	for ( i0=0; i0< tablesLength; i0++ )
	      originalBitmapLocationTable[ i0 ] = (UInt16)fileGetWord( inFilePtr );
	for ( i0=0; i0< (tablesLength-1); i0++ )
	      originalOffsetWidthTable[ i0 ] = (UInt16)fileGetWord( inFilePtr );

	// Separate font parts.
	for( i1=0; i1 < fontHeaderV2.densityCount; i1++ )
	   {
		fileSeek( inFilePtr, (UInt32)desiredFontHeaderLength+desiredDensityEntryLength*i1 );
		if ( headerFormatV5 )
		  getFontDensityPartV5( inFilePtr, &fontDensity );
		else
		  getFontDensityPart( inFilePtr, &fontDensity );

		tmpFilePtr = tmpfile();
		if ( tmpFilePtr==NULL )
	   	   {
			printf("\nError: can not create temporary file. Program aborted.\n\00");
			errorExit( errorFileOperation );
	   	   };

		fillAreaByByte( outFileNameStr, 0, dosFileNameLength );
		printf("\nWrite " );

		switch( fontDensity.density )
		{
		  case oneAndAHalfDensityValue:
			resizeByPercent=oneAndAHalfDensityPercent;
			bitsInRow=24;
			sprintf( outFileNameStr, "%s%04x.bin", "NFOH", resourceID );
			printf("One and a Half");
		  break;

		  case doubleDensityValue:
			resizeByPercent=doubleDensityPercent;
			bitsInRow=32;
			sprintf( outFileNameStr, "%s%04x.bin", "NFDD", resourceID );
			printf("Double");
		  break;

		  case tripleDensityValue:
			resizeByPercent=tripleDensityPercent;
			bitsInRow=48;
			sprintf( outFileNameStr, "%s%04x.bin", "NFTD", resourceID );
			printf("Triple");
		  break;

		  case quadrupleDensityValue:
			bitsInRow=64;
			resizeByPercent=quadrupleDensityPercent;
			sprintf( outFileNameStr, "%s%04x.bin", "NFQD", resourceID );
			printf("Quadruple");
		  break;

		  case singleDensityValue:
		  default:
			resizeByPercent=singleDensityPercent;
			bitsInRow=16;
			if ( fontDensity.density )
			  {
			   sprintf( outFileNameStr, "%s%04x.bin", "NFSD", resourceID );
			   printf("Single");
			  }
			else
			  {
			   sprintf( outFileNameStr, "%s%04x.bin", "UNKN", resourceID );
			   printf("Unknown");
			  };
		  break;
		 };

		outFileNamePtr=outFileNameStr;
		printf(" density font part to '%s' file...", outFileNamePtr );

		// Computate font tables for the separated density data.
		totalGlyphWidth=0;
		for ( i0=0; i0 < tablesLength-1; i0++ )
		  {
			correctedOffsetWidthTable[i0]=originalBitmapLocationTable[i0+1]-originalBitmapLocationTable[i0];

			if ( (originalOffsetWidthTable[i0]&0xFF) < correctedOffsetWidthTable[i0] )
				correctedOffsetWidthTable[i0] = originalOffsetWidthTable[i0];

			if ( (correctedOffsetWidthTable[i0]&0xFF)==0 || (correctedOffsetWidthTable[i0]&0xFF)==0xFF )
				correctedOffsetWidthTable[i0]=0xFFFF;
			else
			   {
			     correctedOffsetWidthTable[i0]=scaleValueByPercent( correctedOffsetWidthTable[i0], resizeByPercent, useCeilingWhenScaleWidth );
			     totalGlyphWidth+=correctedOffsetWidthTable[i0];
			   };
		  };

		correctedOffsetWidthTable[ tablesLength-1 ]=0xFFFF;
		correctedBitmapLocationTable[ 0 ]=0;
		for ( i0=1; i0 < tablesLength; i0++ )
		    correctedBitmapLocationTable[i0]=correctedBitmapLocationTable[i0-1]+((correctedOffsetWidthTable[i0-1]&0xFF)!=0xFF ? (correctedOffsetWidthTable[ i0-1 ]&0xFF):0);

		fontHeaderV1.fontType=0x9000;  // Simply NFNT 9000 font.
		fontHeaderV1.firstGlyph = fontHeaderV2.firstGlyph;
		fontHeaderV1.lastGlyph = fontHeaderV2.lastGlyph;

		fontHeaderV1.maxWidth = scaleValueByPercent( fontHeaderV2.maxWidth, resizeByPercent, useCeilingWhenScaleWidth );
		fontHeaderV1.kernMax = scaleValueByPercent( fontHeaderV2.kernMax, resizeByPercent, useCeilingWhenScaleWidth );
		fontHeaderV1.nDescent = scaleValueByPercent( fontHeaderV2.nDescent, resizeByPercent, useCeilingWhenScaleHeight );

		fontHeaderV1.fRectWidth = scaleValueByPercent( fontHeaderV2.fRectWidth, resizeByPercent,  useCeilingWhenScaleWidth );
		fontHeaderV1.fRectHeight = scaleValueByPercent( fontHeaderV2.fRectHeight, resizeByPercent, useCeilingWhenScaleHeight );

		if ( fontDensity.density==oneAndAHalfDensityValue ) // Specially for One And a Half Density.
		 fontHeaderV1.rowWords = ( (UInt32)totalGlyphWidth+15l )/16l;
		else
		 fontHeaderV1.rowWords = ( scaleValueByPercent( fontHeaderV2.rowWords*2, resizeByPercent,  useCeilingWhenScaleWidth )+1)/2;

		fontHeaderV1.owTLoc = fontHeaderV1.rowWords * (UInt32)fontHeaderV1.fRectHeight+(UInt32)tablesLength+5l;

		fontHeaderV1.ascent = scaleValueByPercent( fontHeaderV2.ascent, resizeByPercent, useCeilingWhenScaleHeight );
		fontHeaderV1.descent = scaleValueByPercent( fontHeaderV2.descent, resizeByPercent, useCeilingWhenScaleHeight );
		fontHeaderV1.leading = scaleValueByPercent( fontHeaderV2.leading, resizeByPercent, useCeilingWhenScaleHeight );

        	fileSeek( tmpFilePtr, 0l ); // Seek file at the start position.
		putFontHeaderV1( &fontHeaderV1, tmpFilePtr );

		fileSeek( inFilePtr, (UInt32)fontDensity.glyphBitsOffset );
		fileSeek( tmpFilePtr, (UInt32)fontTag1Length );

		// Transfer glyph images. Specially for One And a Half Density.
		if ( fontDensity.density==oneAndAHalfDensityValue )
		  {
			UInt8	dotsToShiftLeft=0, dotsToShiftRight=0;
			UInt8	glyphLeftByte=0, glyphRigthByte=0;
			UInt16	rowCounter, columnCounter;
			UInt32	filePos, rowBytes=0;
			UInt8	tileBytes[ maxGlyphColumnHeight ];

     			fillAreaByByte( glyphDataPtr, 0xDB, bufferLength );
			for( columnCounter=0; columnCounter < fontHeaderV1.rowWords; columnCounter++ )
				{
	 			  if ( fwrite( glyphDataPtr, 1, (UInt32)( fontHeaderV1.fRectHeight * 2 ), tmpFilePtr )
	      				!=(UInt32)( fontHeaderV1.fRectHeight * 2 ) )
					{
		  			  puts("\nError: Can not write to file. Program aborted.\00");
		  			  errorExit( errorFileOperation );
					};
				};

			// Get number of bytes per row in source files.
			rowBytes=( scaleValueByPercent( fontHeaderV2.rowWords*2, resizeByPercent, useCeilingWhenScaleWidth )+1)&~1;

			for( i0=0; i0 < (tablesLength-1); i0++ )
			  {
				fillAreaByByte( glyphDataPtr, 0, bufferLength );

				glyphDataPtr->width = ((correctedOffsetWidthTable[i0]&0xFF)!=0xFF)? correctedOffsetWidthTable[i0]&0xFF : 0;
				if ( glyphDataPtr->width <= 0 ) continue;

				glyphDataPtr->height = fontHeaderV1.fRectHeight;
	 			glyphDataPtr->bufferPtr = (UInt8 *)glyphDataPtr+sizeof(glyphBufferType)+1;

				glyphDataPtr->widthInBytes=(glyphDataPtr->width+7)>>3;
				dotsToShiftLeft=scaleValueByPercent( originalBitmapLocationTable[ i0 ], resizeByPercent, useCeilingWhenScaleWidth )&0x7;
				dotsToShiftRight=(8-dotsToShiftLeft)&0x7;

				// Read glyph from source file.
				for( rowCounter=0; rowCounter < glyphDataPtr->height; rowCounter++ )
				 for( columnCounter=0; columnCounter < glyphDataPtr->widthInBytes; columnCounter++ )
				    {
					(UInt32)filePos=(UInt32)fontDensity.glyphBitsOffset+(UInt32)( scaleValueByPercent( originalBitmapLocationTable[ i0 ], resizeByPercent, useCeilingWhenScaleWidth )>>3 )+(UInt32)columnCounter+(UInt32)rowCounter*(UInt32)rowBytes;
				 	fileSeek( inFilePtr, (UInt32)filePos );

		  			glyphLeftByte = fileGetByteWFName( inFilePtr, argv[i] );
		  			glyphRigthByte = dotsToShiftLeft!=0 ? fileGetByteWFName( inFilePtr, argv[i] ) : 0;

		  			glyphDataPtr->bufferPtr[ (rowCounter * (UInt32)glyphDataPtr->widthInBytes)+columnCounter ] |=
						( (glyphLeftByte<<dotsToShiftLeft)&0xFF) | ( (glyphRigthByte>>dotsToShiftRight)&0xFF );
				    };

				// Write glyph image to the output file.
				if ( (correctedBitmapLocationTable[ i0 ]&0x07)!=0 )
				    { 
					UInt8	shiftRow[ rowLength+1 ];
		  			UInt8	shiftRowWidth, shiftRowWidthInBytes;

					dotsToShiftRight = correctedBitmapLocationTable[ i0 ]&0x7;
					dotsToShiftLeft = (8-dotsToShiftRight)&0x7;

 	 	  			shiftRowWidth = glyphDataPtr->width + dotsToShiftRight;
					shiftRowWidthInBytes = (shiftRowWidth+7) >> 3;

					for( rowCounter=0; rowCounter < glyphDataPtr->height; rowCounter++ )
					  {
					    for( columnCounter = 0; columnCounter < shiftRowWidthInBytes; columnCounter++ )
						 if ( columnCounter==0 )
						   shiftRow[ columnCounter ] = ( tileBytes[ rowCounter ] & ((0xFF) << dotsToShiftLeft )) | (glyphDataPtr->bufferPtr[ rowCounter * glyphDataPtr->widthInBytes ] >> dotsToShiftRight );
			   			 else
			   			   shiftRow[ columnCounter ] = ( glyphDataPtr->bufferPtr[ columnCounter-1 + (rowCounter * glyphDataPtr->widthInBytes) ] << dotsToShiftLeft ) |
							    ( ( columnCounter < glyphDataPtr->widthInBytes ) ?
		                               		    ( glyphDataPtr->bufferPtr[ columnCounter + (rowCounter * glyphDataPtr->widthInBytes) ] >> dotsToShiftRight ) : 0 );

	      				    proceedBinaryDataFileOut( tmpFilePtr, &shiftRow[ 0 ],
	   			        			  (UInt32)fontTag1Length + (UInt32)( correctedBitmapLocationTable[ i0 ]>>3 ) + (UInt32)fontHeaderV1.rowWords * 2l * (UInt32)rowCounter,
           			        			  shiftRowWidthInBytes );

					    // Accomulate last bytes of shifted row to use in the next glyph write;
	      				    tileBytes[ rowCounter ] = shiftRow[ (shiftRowWidthInBytes-1) ];
			  		  };
				    }
				 else
	     			   for( rowCounter=0; rowCounter < glyphDataPtr->height; rowCounter++ )
	       			     {
	        			proceedBinaryDataFileOut( tmpFilePtr, &glyphDataPtr->bufferPtr[ rowCounter * glyphDataPtr->widthInBytes ],
	   			        			  (UInt32)fontTag1Length + (UInt32)( correctedBitmapLocationTable[ i0 ]>>3 ) + (UInt32)fontHeaderV1.rowWords * 2l * (UInt32)rowCounter,
            			        			  glyphDataPtr->widthInBytes );
					// Store last byte of glyphs row;
	        			tileBytes[ rowCounter ] = glyphDataPtr->bufferPtr[ (glyphDataPtr->widthInBytes-1) + rowCounter * glyphDataPtr->widthInBytes ];
	       			     };
			     };

		  } // End for One and a Half density treatment.
		else // For densities other than one and a half simply copy image data.
		for( i0=0; i0 < fontHeaderV1.fRectHeight*(UInt32)fontHeaderV1.rowWords; i0++ )
		  {
		    wordValue = fileGetWordWFName( inFilePtr, argv[i] );
		    filePutWord( tmpFilePtr, wordValue );
		  };

		// Seek and write to the font table position.
		(UInt32)imagePartLength=(UInt32)fontHeaderV1.fRectHeight*(UInt32)fontHeaderV1.rowWords*2l;
		fileSeek( tmpFilePtr, (UInt32)fontTag1Length + (UInt32)imagePartLength );
		for( i0=0; i0<tablesLength; i0++ ) filePutWord( tmpFilePtr, correctedBitmapLocationTable[i0] );
		for( i0=0; i0<tablesLength; i0++ ) filePutWord( tmpFilePtr, correctedOffsetWidthTable[i0] );

		// Complete output file.
		outFilePtr=openFile( outFileNamePtr, "wb+");
		if ( outFilePtr==NULL )
		  {
			printf("\nError: can not open file to write '%s'. Program aborted.", outFileNamePtr );
			errorExit( errorFileOperation );
		  };
		(unsigned long)fileSize = (unsigned long)fileLength( tmpFilePtr ); // Get size of opened file;
		fileSeek( tmpFilePtr, 0l );
		proceedBinaryDataFileTransfer( tmpFilePtr, outFilePtr, 0l, (unsigned long)fileSize );
		fclose( outFilePtr );
		fclose( tmpFilePtr );
		printf(" Done.");
	   }; // End of separating loop.
	 fclose( inFilePtr ); // Close input file.
       };
   }; // End of command line arguments counter;

 if ( argc==1 ) 
	usage(); // Display help list;

 return returnStatus; 
};

// Display user's help.
static void usage( void )
{
  puts("\nUsage ([] means optional):");
  puts("   flushfv2 [flags] <fontres.bin>...");
  puts("               ... fontres.bin is a Palm OS(tm) binary nfnt font resource.\n");
  puts("Flags are:\n");
  puts("-n<value>     value to specify base resource id number for output files.");
  puts("-h, -?	      help screen.");
};
