// Copyright (c) 2005 Alexander S. Ovcharenko, SymbioForge.
//
// FntScale - Scale font glyph images of binary Palm font resource by percent.
//
// This program is free software; you can redistribute it and/or modify it
// under the terms of the GNU General Public License as published by the
// FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
// opinion) any later version.
//
// To obtain a complete GNU License from internet use URL http://www.fsf.org
// -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
// gcc fntscale.c toollib.c -o fntscale
//
//  Where GCC is a GNU C compiler or any other suitable to your computer 
// operation system;

#include <stdio.h>
#include "toollib.h"


// A local structure with program arguments data.
struct {
	Boolean			helpScreen;
	Boolean			verboseOut;
 	char			nameOfResourceFile[ filePathLength+1 ];
 	char			nameOfOutResourceFile[ filePathLength+1 ];
	Boolean			scaleValueSetFlag;
	Boolean			useBackScaleFlag;
	short			scaleValue;		// Percent for scale;
	Boolean			isCorrect;
       } argsData;

static void	usage( void );	// Display help screen.

main( int argc, char * argv[] )
 { short		i, arglen, resNameLength;
   unsigned long	fileSize;

   puts("FntScale - Scale font glyph images of PalmOS font resource by percent value." );
   puts("Version 1.3 - (c) 2005, SymbioForge (http://www.symbioforge.com/feii/)." );

   fillAreaByByte( &argsData, 0, sizeof( argsData ) );

   for( i=1; i < argc; i++ )
    {
     arglen=strlen( argv[i] ); // Get length of this argument;

     if ( (arglen>=1) && (argv[i][0]=='-' || argv[i][0]=='/' || argv[i][0]=='%') )
        {   // Check for help request.

	    if ( (arglen>=2) && ( argv[i][1]=='%' || argv[i][0]=='%' ) )
	       { UInt16	percentValue;
	         if ( argsData.scaleValueSetFlag )
		    {
		      puts( "\nError: scale value already set. Program stopped.\00" );
  		      errorExit( errorExitValue );
		    };

	         percentValue= (argv[i][0]=='%') ? atoi(&argv[i][1]):atoi(&argv[i][2]) ;
	         if ( percentValue<=0 || percentValue>1000 )
		    {
		      puts( "\nError: invalid percent value for scale. Program stopped.\00" );
  		      errorExit( errorExitValue );
		    };

	         argsData.scaleValue = percentValue;
	         argsData.scaleValueSetFlag = true; // Set scale value flag to true;
	         argsData.useBackScaleFlag = false; // Set scale value flag to true;
               }
            else
	    if ( (arglen==4) && tolower(argv[i][2])=='t' )
	       { char densityCase=' ';

	          argsData.scaleValueSetFlag = true; // Set scale value flag to true;
		  if ( tolower(argv[i][1])=='s') // to enlarge image of glyphs.
		      {
			  densityCase=tolower( argv[i][3] );
	          	  argsData.useBackScaleFlag = false;
		      }
		  else
		  if ( tolower(argv[i][3])=='s') // to reduce glyphs image dimension.
		      {
			  densityCase=tolower(argv[i][1]);
	          	  argsData.useBackScaleFlag = true; // Set scale value flag to true;
		      }

		 switch( densityCase )
		      {
			  case 'o': // From one and a half density.
	         		argsData.scaleValue = oneAndAHalfDensityPercent;
			  break;

			  case 'd': // Double density.
	         		argsData.scaleValue = doubleDensityPercent;
			  break;

			  case 't': // Triple density.
	         		argsData.scaleValue = tripleDensityPercent;
			  break;

			  case 'q': // Quadruple density.
	         		argsData.scaleValue = quadrupleDensityPercent;
			  break;

			  default:
	         	    	argsData.scaleValueSetFlag = false; // Set scale value flag to true;
			  break;
		      };
	       }
	    else
 	    if ( (arglen==2) && (tolower(argv[i][1])=='h') || argv[i][1]=='?')
		argsData.helpScreen=1;
	    else
	    if ( (arglen==2) && (tolower(argv[i][1])=='v') )
		argsData.verboseOut=1; // Do verbose output for scaled glyphs;
	    else  // Check for out file option set;
	    if ( (arglen>=3) && ( tolower(argv[i][1])=='o') )
		{ char	* outNamePtr;

		  // Extract file name of font for input.
		  outNamePtr = arglen > 3 ? &argv[i][2] : NULL;
		  resNameLength = strlen( outNamePtr );

		  if ( strlen(argsData.nameOfOutResourceFile)!=0 )
		     {
     		 	puts( "\nError: output font file already set. Program stopped.\00" );
  		 	errorExit( errorExitValue );
		     };

		  // Display an error when output font resource is too long.
	    	  if ( resNameLength > filePathLength ) 
		     {
     		 	puts( "\nError: output font file name is too long. Program stopped.\00" );
  		 	errorExit( errorExitValue );
		     };

                  strncpy( argsData.nameOfOutResourceFile, outNamePtr, resNameLength );
		  argsData.nameOfOutResourceFile[ resNameLength ] = 0;
		 }
	      else
		 {
			puts( "\nError: invalid argument used. Program stopped.\00" );
			errorExit( errorExitValue );
		 };
         } // When no flag '-' or '+' assume this is a resource file name to scale.
	else
	 { 
	    resNameLength = strlen( argv[ i ] );
	    if ( strlen( argsData.nameOfResourceFile )!=0 )
		{
		 puts( "\nError: font resource file already set. Program stopped.\00" );
  		 errorExit( errorExitValue );
		};

	    if ( resNameLength > filePathLength )
		{
     		 puts( "\nError: font resource name is too long. Program stopped.\00" );
  		 errorExit( errorExitValue );
		};

	    strncpy( argsData.nameOfResourceFile, argv[ i ], resNameLength );
	    argsData.nameOfResourceFile[ resNameLength ] = 0;
        };
   }; // End of argument counting;

// Check for correct argument data.
argsData.isCorrect = false;
if ( (argsData.scaleValueSetFlag) || (strlen( argsData.nameOfResourceFile )>0) || (argsData.verboseOut) || (strlen( argsData.nameOfOutResourceFile )!=0) )
  {
    if ( strlen( argsData.nameOfResourceFile )==0 )
	{
	  puts( "\nError: font resource name is not set. Program stopped.\00" );
	  errorExit( errorExitValue );
	}
    else
    if ( !argsData.scaleValueSetFlag )
	{
          puts( "\nError: scale value is not set. Program stopped.\00" );
	  errorExit( errorExitValue );
	};
    argsData.isCorrect = true;
   };

if ( strlen(argsData.nameOfOutResourceFile)==0 ) // Create name for output file.
	strcpy( argsData.nameOfOutResourceFile, argsData.nameOfResourceFile );

// Scale font resource;
if ( argsData.isCorrect )
   { FILE		* resFilePtr = NULL;
     FILE		* fileOutPtr = NULL;
     FILE 		* tmpFilePtr = NULL;
     Int8		fontVersion = 0;

     UInt8 *		bufferPtr=NULL;
     glyphBufferType	* originalGlyphPtr=NULL;  // Glyph image retreived from database.
     glyphBufferType	* modifiedGlyphPtr=NULL;  // Scaled glyph image.

     FontType		fontHeader, scaledFontHeader;
     UInt16		i, tablesLength, glyphWidth, glyphKern;
     UInt16		bitmapLocationTable[ maxTableElements ], modifiedBitmapLocationTable[ maxTableElements ]; // Array of offsets to glyph bitmap images.
     UInt16		offsetWidthTable[ maxTableElements ], modifiedOffsetWidthTable[ maxTableElements ]; // Array of glyph width table;
     UInt32		fntHeaderPos, fntGlyphBitmapPos, fntOffsetWidthTblPos, fntBitmapLocationPos, filePos;
     UInt32		fontImageWidth=0l, offsetWidthTableLocation=0l, fileSize=0;

     UInt16		glyphBitmapOffset, glyphBitmapOffsetInBytes;
     UInt16		glyphDestWidth, glyphDestHeight, glyphDestKern;
     UInt8		dotsToShiftRight=0, dotsToShiftLeft=0, rowCounter; //, bitmapWidthInBytes;
     UInt16		columnCounter;
     UInt8		glyphLeftByte, glyphRigthByte;

     UInt8		originalGlyphData[ bufferLength ];
     UInt8		modifiedGlyphData[ bufferLength ];
     UInt8		tileBytes[ maxGlyphColumnHeight ];
     UInt8		dotsToShift, widthCount;

     resFilePtr = openFile( argsData.nameOfResourceFile, "rb"); 
     (UInt32)fileSize=(UInt32)fileLength( resFilePtr );

     if ( resFilePtr==NULL )
	{
	  printf("\nError: can not open file '%s'. Program aborted.\n\00", argsData.nameOfResourceFile );
	  errorExit( errorFileOperation );
	};

     // Check font resource for consistency.
     fontVersion = getFontResourceType( resFilePtr, true );
     if ( fontVersion==0 )
	{
	  printf("\nError: file '%s' is not a binary font resource. Program stopped.\n\00", argsData.nameOfResourceFile );
	  errorExit( errorFileOperation );
	}
     else
     if ( fontVersion!=1 )
	{
	  printf("\nError: font resource '%s' is not type version 1. Program aborted.\n\00", argsData.nameOfResourceFile );
	  errorExit( errorFileOperation );
	};

     tmpFilePtr = tmpfile();
     if ( tmpFilePtr==NULL )
	{
          printf("\nError: can not open temporary file. Program aborted.\n\00");
	  errorExit( errorFileOperation );
	};

     originalGlyphPtr = (glyphBufferType *)&originalGlyphData[ 0 ];
     modifiedGlyphPtr = (glyphBufferType *)&modifiedGlyphData[ 0 ];

     // get font header.
     fileSeek( resFilePtr, 0l ); // Seek file at the start position.
     fillAreaByByte( &scaledFontHeader, 0, sizeof( FontType ) );

     getFontHeaderV1WFName( resFilePtr, &fontHeader, argsData.nameOfResourceFile ); // 

     scaledFontHeader.fontType   	= 0x9000;
     scaledFontHeader.firstGlyph 	= fontHeader.firstGlyph; //
     scaledFontHeader.lastGlyph  	= fontHeader.lastGlyph;  //

     if ( argsData.useBackScaleFlag )
	{
     	 scaledFontHeader.maxWidth	= scaleBackValueByPercent( fontHeader.maxWidth, argsData.scaleValue, useCeilingWhenScaleWidth );
     	 scaledFontHeader.fRectWidth	= scaleBackValueByPercent( fontHeader.fRectWidth, argsData.scaleValue, useCeilingWhenScaleWidth );
     	 scaledFontHeader.descent	= scaleBackValueByPercent( fontHeader.descent, argsData.scaleValue, useCeilingWhenScaleHeight );
     	 scaledFontHeader.ascent	= scaleBackValueByPercent( fontHeader.ascent, argsData.scaleValue, useCeilingWhenScaleHeight );
     	 scaledFontHeader.leading 	= scaleBackValueByPercent( fontHeader.leading, argsData.scaleValue, useCeilingWhenScaleHeight );
     	 scaledFontHeader.kernMax       = scaleBackValueByPercent( fontHeader.kernMax, argsData.scaleValue, useCeilingWhenScaleWidth );
     	 scaledFontHeader.fRectHeight	= scaleBackValueByPercent( fontHeader.fRectHeight, argsData.scaleValue, useCeilingWhenScaleHeight );
	}
      else
	{
     	 scaledFontHeader.maxWidth	= scaleValueByPercent( fontHeader.maxWidth, argsData.scaleValue, useCeilingWhenScaleWidth );
     	 scaledFontHeader.fRectWidth	= scaleValueByPercent( fontHeader.fRectWidth, argsData.scaleValue, useCeilingWhenScaleWidth );
     	 scaledFontHeader.descent	= scaleValueByPercent( fontHeader.descent, argsData.scaleValue, useCeilingWhenScaleHeight );
     	 scaledFontHeader.ascent	= scaleValueByPercent( fontHeader.ascent, argsData.scaleValue, useCeilingWhenScaleHeight );
     	 scaledFontHeader.leading 	= scaleValueByPercent( fontHeader.leading, argsData.scaleValue, useCeilingWhenScaleHeight );
     	 scaledFontHeader.kernMax       = scaleValueByPercent( fontHeader.kernMax, argsData.scaleValue, useCeilingWhenScaleWidth );
     	 scaledFontHeader.fRectHeight	= scaleValueByPercent( fontHeader.fRectHeight, argsData.scaleValue, useCeilingWhenScaleHeight );
	};

     if ( scaledFontHeader.fRectHeight > 127 )
	{
	  printf("\nError: height of scaled glyph exceeds 127 points. Program stopped.\n\00");
  	  errorExit( errorExitValue );
     	};

     tablesLength = (fontHeader.lastGlyph-fontHeader.firstGlyph)+3; // Lenght of font tables in words;
     (UInt32)fntGlyphBitmapPos=(UInt32)fontTag1Length;
     (UInt32)fntBitmapLocationPos=(UInt32)fntGlyphBitmapPos+(UInt32)fontHeader.rowWords*2*(UInt32)fontHeader.fRectHeight;
     (UInt32)fntOffsetWidthTblPos=(UInt32)fntBitmapLocationPos+(UInt32)tablesLength*2;

     // printf("Bitmap location pos: %lu\n", fntBitmapLocationPos );

     // Read font tables;
     fileSeek( resFilePtr, (UInt32)fntBitmapLocationPos ); // Seek file at the start position.
     for( i=0; i < tablesLength; i++ )
	bitmapLocationTable[ i ] = fileGetWordWFName( resFilePtr, argsData.nameOfResourceFile );

     fileSeek( resFilePtr, (UInt32)fntOffsetWidthTblPos ); // Seek file at the start position.
     for( i=0; i < tablesLength; i++ )
      {
	if ( ftell(resFilePtr) < fileSize )
		offsetWidthTable[ i ] = fileGetWordWFName( resFilePtr, argsData.nameOfResourceFile );
	else
		offsetWidthTable[ i ] = 0xFFFF;
      };
     // Scale font tables' elements by percent;
     (UInt32)fontImageWidth=0l;
     modifiedBitmapLocationTable[ 0 ] = 0;
     modifiedOffsetWidthTable[ tablesLength-1 ] = offsetWidthTable[ tablesLength-1 ];

     fillAreaByByte( &tileBytes, 0, maxGlyphColumnHeight ); // Initialize tileBytes array.
     for( i=0; i < tablesLength-1; i++ ) // Scale every glyph in the font;
	{ 
	 // Get width from the difference of bitmap locations.
 	 glyphWidth = bitmapLocationTable[i+1]-bitmapLocationTable[i];

	 if ( ((offsetWidthTable[i]&0xFF)!=0) && ((offsetWidthTable[i]&0xFF)!=0xFF) && ((offsetWidthTable[i]&0xFF<glyphWidth)) )
	   glyphWidth=offsetWidthTable[i]&0xFF;

	 glyphKern = offsetWidthTable[i]>>8; // Get glyph kern.

	 if ( argsData.useBackScaleFlag )
	    {
	 	glyphDestWidth = scaleBackValueByPercent( glyphWidth, argsData.scaleValue, useCeilingWhenScaleWidth );
	 	glyphDestKern = scaleBackValueByPercent( glyphKern, argsData.scaleValue, useCeilingWhenScaleWidth );
	    }
	 else
	    {
	 	glyphDestWidth = scaleValueByPercent( glyphWidth, argsData.scaleValue, useCeilingWhenScaleWidth );
	 	glyphDestKern = scaleValueByPercent( glyphKern, argsData.scaleValue, useCeilingWhenScaleWidth );
	    };

     	 if ( glyphDestWidth > maxGlyphColumnHeight )
	    {		
	 		printf("\nError: width of scaled glyph exceeds %u points. Program aborted.\n\00", maxGlyphColumnHeight );
	 		errorExit( errorExitValue );
	    };
	 (UInt32)fontImageWidth += glyphDestWidth;
	 modifiedBitmapLocationTable[ i+1 ] = modifiedBitmapLocationTable[ i ] + glyphDestWidth;
	 modifiedOffsetWidthTable[ i ] = (glyphDestWidth == 0) ? (0xFFFF) : (glyphDestKern<<8) | (glyphDestWidth&0xFF); // Array of glyph width table;
	};

     // Continue to computate font header.
     scaledFontHeader.rowWords= (fontImageWidth+15)/16;	// Calculate amount of words in row.
     (UInt32)offsetWidthTableLocation  = (UInt32)scaledFontHeader.rowWords * (UInt32)scaledFontHeader.fRectHeight+(UInt32)tablesLength+(UInt32)5;
     (UInt16)scaledFontHeader.owTLoc   = (offsetWidthTableLocation & 0xFFFF); // Get low word of offset to WidthTableLocations.
     (UInt16)scaledFontHeader.nDescent = (UInt32)offsetWidthTableLocation > (UInt32)0xFFFF ? (UInt32)offsetWidthTableLocation >>16 : ( fontHeader.nDescent==0 ? 0 : -1 );

     if ( argsData.verboseOut ) // display scaled font header in text PilRC mode.
	{
	 puts("");
	 printf("fontType            %hu\n", scaledFontHeader.fontType);
	 printf("fRectHeight         %hu\n", scaledFontHeader.fRectHeight);
	 printf("ascent              %hu\n", scaledFontHeader.ascent);
	 printf("descent             %hu\n", scaledFontHeader.descent);
	 puts("");
	}
     else
	printf("\nScaling the font...");

     // Proceed header, empty glyphs, tables write to temporary file;
     fileSeek( tmpFilePtr, 0l ); // Seek temporary file at the start position.
     // Put font header;
     putFontHeaderV1( (FontType *)&scaledFontHeader, tmpFilePtr );

     // OS may return error when no space at place it seeks, so
     // make empty glyph areas in file. It will be filled by images later.
     fillAreaByByte( modifiedGlyphPtr, 0, bufferLength );
     for( columnCounter=0; columnCounter < scaledFontHeader.rowWords; columnCounter++ )
	{
	 if ( fwrite( modifiedGlyphPtr, 1, (unsigned long)( scaledFontHeader.fRectHeight * 2 ), tmpFilePtr )
	      !=(unsigned long)( scaledFontHeader.fRectHeight * 2 ) )
		{
		  puts("\nError: Can not write to file. Program aborted.\00");
		  errorExit( errorFileOperation );
		};
	};

     // Add tables to file.
     for ( i=0; i < tablesLength; i++ )
	 filePutWord( tmpFilePtr, modifiedBitmapLocationTable[ i ] );
     for ( i=0; i < tablesLength; i++ )
	 filePutWord( tmpFilePtr, modifiedOffsetWidthTable[ i ] );

     for( i=0; i<=(fontHeader.lastGlyph-fontHeader.firstGlyph)+1; i++ ) // Scale every glyph in the font;
	{
	 // Create glyph structure.
	 fillAreaByByte( originalGlyphPtr, 0, bufferLength );

	 originalGlyphPtr->width=bitmapLocationTable[i+1]-bitmapLocationTable[i];
	 if ( originalGlyphPtr->width <= 0 ) continue;	// Do not iterate this loop.


	 originalGlyphPtr->widthInBytes = ( originalGlyphPtr->width+7)>>3;
	 originalGlyphPtr->height = fontHeader.fRectHeight;
	 originalGlyphPtr->bufferPtr = (UInt8 *)originalGlyphPtr+sizeof(glyphBufferType)+1;
	 dotsToShiftLeft = bitmapLocationTable[ i ] & 7; // How many bits to shift right when add first byte.
	 dotsToShiftRight = ( 8-dotsToShiftLeft ) & 7;   // How many bits to shift left at adding next byte;

	 // Load glyph from file;
	 if ( originalGlyphPtr!=NULL )
	 for( rowCounter=0; rowCounter < originalGlyphPtr->height; rowCounter++ )
	   for( columnCounter=0; columnCounter < originalGlyphPtr->widthInBytes; columnCounter++ )
		{
		  // (UInt32)filePos=(UInt32)fntGlyphBitmapPos+(UInt32)glyphBitmapOffsetInBytes+(UInt32)columnCounter+(UInt32)rowCounter*(UInt32)fontHeader.rowWords*(UInt32)2;
		  (UInt32)filePos=(UInt32)fntGlyphBitmapPos+(UInt32)(bitmapLocationTable[i]>>3)+(UInt32)columnCounter+(UInt32)rowCounter*(UInt32)fontHeader.rowWords*(UInt32)2;
		  fileSeek( resFilePtr, (UInt32)filePos );

		  // glyphLeftByte = fileGetByte( resFilePtr );
		  glyphLeftByte = fileGetByteWFName( resFilePtr, argsData.nameOfResourceFile );
		  // glyphRigthByte = dotsToShiftLeft!=0 ? fileGetByte( resFilePtr ) : 0;
		  glyphRigthByte = dotsToShiftLeft!=0 ? fileGetByteWFName( resFilePtr, argsData.nameOfResourceFile ) : 0;
		  originalGlyphPtr->bufferPtr[ (rowCounter * (UInt32)originalGlyphPtr->widthInBytes)+columnCounter ] |=
			( (glyphLeftByte<<dotsToShiftLeft)&0xFF) | ( (glyphRigthByte>>dotsToShiftRight)&0xFF );
		};

	 // Computate width and height of destination glyph for scale;
	 glyphDestWidth = modifiedOffsetWidthTable[ i ]; // Array of glyph width table.

	 // Create space for destination glyph;
	 fillAreaByByte( modifiedGlyphPtr, 0, bufferLength );
	 // memset( modifiedGlyphPtr, 0, bufferLength );
	 modifiedGlyphPtr->width = glyphDestWidth;
	 modifiedGlyphPtr->widthInBytes = ( modifiedGlyphPtr->width+7 )>>3;
	 modifiedGlyphPtr->height = scaledFontHeader.fRectHeight;
	 modifiedGlyphPtr->bufferPtr = (UInt8 *)modifiedGlyphPtr+sizeof(glyphBufferType);
	 glyphBufferScale( modifiedGlyphPtr, originalGlyphPtr ); // Scale the glyph.


	 // FontType fontHeader, scaledFontHeader;
	 if ( argsData.verboseOut && modifiedGlyphPtr!=0 ) // display glyphs body text PilRC mode.
	   {
		if ( i+fontHeader.firstGlyph > fontHeader.lastGlyph )
			printf("GLYPH -1 // Missing character. ");
		else
			printf("GLYPH %u // ", i + fontHeader.firstGlyph );

		printf( "width: %u\n", modifiedGlyphPtr->width );

		if ( modifiedGlyphPtr!=NULL )
	 	  for( rowCounter=0; rowCounter < scaledFontHeader.fRectHeight; rowCounter++ )
	    		{
	  		 for( columnCounter=0; columnCounter < modifiedGlyphPtr->width; columnCounter++ )
		    	    printf( "%s", ( ( modifiedGlyphPtr->bufferPtr[ (columnCounter>>3)+(rowCounter*modifiedGlyphPtr->widthInBytes) ] & ( 0x80>>( columnCounter & 0x07 ) ) ) !=0 ) ? "#" : "-" );
	       		  puts("");
			};
     	  	puts("");
	   };

	 // Shift glyph right when add last bits of previous glyph.
	 if ( modifiedGlyphPtr->width!=0 )
	   {

            if ( (modifiedBitmapLocationTable[ i ]&0x07)!=0 )
	      {	
		UInt8	shiftRow[ rowLength+1 ];
		UInt8	shiftRowWidth, shiftRowWidthInBytes;

		// Add last part of previous glyph which does not fit to byte;
		dotsToShiftRight = modifiedBitmapLocationTable[ i ]&0x07;
		dotsToShiftLeft = ( 8-dotsToShiftRight ) & 0x07; // How many bits to shift left at adding next byte;
 	 	shiftRowWidth = modifiedGlyphPtr->width + dotsToShiftRight;
		shiftRowWidthInBytes = (shiftRowWidth+7) >> 3;

		for( rowCounter=0; rowCounter < scaledFontHeader.fRectHeight; rowCounter++ )
		   {
			//fillAreaByByte( &shiftRow, 0, rowLength );
			for( columnCounter = 0; columnCounter < shiftRowWidthInBytes; columnCounter++ )
			  {
			   if ( columnCounter==0 )
				shiftRow[ columnCounter ] = ( tileBytes[ rowCounter ] & ((0xFF) << dotsToShiftLeft )) | (modifiedGlyphPtr->bufferPtr[ rowCounter * modifiedGlyphPtr->widthInBytes ] >> dotsToShiftRight );
			   else
			   	shiftRow[ columnCounter ] = ( modifiedGlyphPtr->bufferPtr[ columnCounter-1 + (rowCounter * modifiedGlyphPtr->widthInBytes) ] << dotsToShiftLeft ) |
							    ( ( columnCounter < modifiedGlyphPtr->widthInBytes ) ?
		                               		      ( modifiedGlyphPtr->bufferPtr[ columnCounter + (rowCounter * modifiedGlyphPtr->widthInBytes) ] >> dotsToShiftRight ) : 0 );
			  };

	      		proceedBinaryDataFileOut( tmpFilePtr, &shiftRow[ 0 ],
	   			        (UInt32)fontTag1Length + (UInt32)(modifiedBitmapLocationTable[ i ]>>3) + (UInt32)scaledFontHeader.rowWords * (UInt32)2 * (UInt32)rowCounter,
            			        shiftRowWidthInBytes );

			// Accumulate last bytes of shifted row to use in the next glyph write;
	      		tileBytes[ rowCounter ] = shiftRow[ (shiftRowWidthInBytes-1) ];
		   };
	      }
           else // Write glyph to the temporary file;
	     for( rowCounter=0; rowCounter < scaledFontHeader.fRectHeight; rowCounter++ )
	       {
	        proceedBinaryDataFileOut( tmpFilePtr, &modifiedGlyphPtr->bufferPtr[ rowCounter * modifiedGlyphPtr->widthInBytes ],
	   			        (UInt32)fontTag1Length + (UInt32)(modifiedBitmapLocationTable[ i ]>>3) + (UInt32)scaledFontHeader.rowWords * (UInt32)2 * (UInt32)rowCounter,
            			        modifiedGlyphPtr->widthInBytes );
	        tileBytes[ rowCounter ] = modifiedGlyphPtr->bufferPtr[ (modifiedGlyphPtr->widthInBytes-1) + rowCounter * modifiedGlyphPtr->widthInBytes ];
	       };
	 };
	};

     fclose( resFilePtr ); // Close binary res file;

     // Move file from temporary file to binary font resource;
     fileOutPtr = openFile( argsData.nameOfOutResourceFile, "wb+" );
     if ( fileOutPtr == NULL )
	{
	  printf("\nError: can not write to file '%s'. Program aborted.\n\00", argsData.nameOfOutResourceFile );
	  errorExit( errorFileOperation );
	};

     (unsigned long)fileSize = (unsigned long)fileLength( tmpFilePtr );
     proceedBinaryDataFileTransfer( tmpFilePtr, fileOutPtr, 0l, (unsigned long)fileSize );

     fclose( fileOutPtr );
     fclose( tmpFilePtr ); // Close temporary res file;

     if ( !argsData.verboseOut ) printf("\nDone.\n");
   };

if ( (argsData.helpScreen) || (!argsData.isCorrect) ) usage();

return;
};

// Display user's help.
static void usage( void )
{
  puts("\nUsage ([] means optional):");
  puts("   fntscale <fontres.bin> [flags]");
  puts("               ... where fontres.bin is a name of font resource to scale.\n");
  puts("Flags are:\n");
  puts("-%<value>     scale percent in the range (1...1000).");
  puts("-StO          scale from SINGLE to ONE AND A HALF densiy.");
  puts("-StD	      scale from SINGLE to DOUBLE density.");
  puts("-StT	      scale from SINGLE to TRIPLE density.");
  puts("-StQ	      scale from SINGLE to QUADRUPLE density.");
  puts("-OtS	      scale back from ONE AND A HALF to SINGLE density.");
  puts("-DtS	      scale back from DOUBLE to SINGLE density.");
  puts("-TtS	      scale back from TRIPLE to SINGLE density.");
  puts("-QtS	      scale back from QUADRUPLE to SINGLE density.\n");

  puts("-h, -?	      this help screen.");
  puts("-o<filename>  alternate name of scaled font file for output.");
  puts("-v	      to verbosely output glyphs in the PilRC text format on screen.\n");
  puts("NOTICE: Font resource will be overwritten when '-o' option not used.");
  puts("      Only one font resource can be treated for the program run.");

  return;
};
