//  Copyright (C) 2008 Alexander S. Ovcharenko, SymbioForge
//
//  fnttopc - utility to convert PalmOS(tm) binary font resources type 1 to
//  the PC Screen fonts either raw or psf.
//
//  This program is free software; you can redistribute it and/or modify 
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
//  gcc fnttopc.c toollib.c -o fnttopc
//
//   Where GCC is a GNU C compiler or other used by your computer operation
//  system;

#include <stdio.h>
#include "toollib.h"

// Define length of font in bytes for the specified dimensions.
#define		dosFirstByte	0
#define		dosLastByte	255

// A local structure with program arguments data.

static		void usage( void );

main( int argc, char *argv[] )
{ int		returnStatus = 0;

  int   		i, i1, i2, i3, arglen;
  int		noFileName = 1;
  int          	fontHeight = 16; //16 pixels by default.

  FILE		* resFilePtr = NULL;  // Filehandle to resource file.
  FILE          	* infoFilePtr = NULL;  // Filehandle of PSF file with unicode information.
  FILE		* fileOutPtr = NULL;  // Filehandle to output file.
  FILE 		* tmpFilePtr = NULL;  // Filehandle to temporary file.
  Int8		fontVersion = 0;
  FontType	fontHeader;

  psfFontV1Type	psf1FontHeader;
  psfFontV2Type	psf2FontHeader;

  UInt16 		tablesLength;
  UInt32		fntGlyphBitmapPos, fntBitmapLocationPos, fntOffsetWidthTblPos;
  UInt8		glyphImageByte;

UInt32		fileSize=0;
UInt32		infoFileUnicodeDataOffset=0l;
  char		nameOfOutResourceFile[ filePathLength+1 ];
  char		nameOfInfoResourceFile[ filePathLength+1 ];
  char		nameOfInResourceFile[ filePathLength+1 ];

  UInt16		bitmapLocationTable[ maxTableElements ];
  UInt16		offsetWidthTable[ maxTableElements ];

  UInt8 *		bufferPtr=NULL;
  glyphBufferType	* originalGlyphPtr=NULL;  // Glyph image retreived from database.

  UInt8		originalGlyphData[ bufferLength ];
  UInt8		tileBytes[ maxGlyphColumnHeight ];
  UInt8		dotsToShift, widthCount;
  nameOfOutResourceFile[0]=0;
  nameOfInResourceFile[0]=0;
  originalGlyphPtr = (glyphBufferType *)&originalGlyphData[ 0 ];

  puts("fnttopc - Convert PalmOS(tm) v1 font resource to Intel(r) x86 PC screen fonts.");
  puts("Version 1.1 (c) 2008, SymbioForge (http://www.symbioforge.com/feii/).");

  for ( i=1; i < argc; i++ )
   { 
    arglen = strlen( argv[i] ); // Get length of this argument;

    // Treat command line options;
    if ( (arglen >= 1 ) && (argv[i][0]=='-' || argv[i][0]=='/') ) // Assume this is a flag;
       {  unsigned int value=0;

          if ( arglen >= 2 ) // Check the flag;
	    {
             char	firstChar, secondChar;

	     firstChar = tolower( argv[ i ][ 1 ] );
	     secondChar = tolower( argv[ i ][ 2 ] );

	     switch( firstChar )
               {
		case 'h':
		case 'H':
		case '?':
			usage();
		break;

		case '8':
			fontHeight=8;
		break;

		case '1':
	     		switch( secondChar )
			 {

				case '4':
					fontHeight=14;
				break;

				case '6':
					fontHeight=16;
				break;

				case '8':
					fontHeight=18;
				break;

				case '9':
					fontHeight=19;
				break;

                		default:
					puts( "\nError: unsupported font height defined. Program aborted.\00" );
					errorExit( errorExitValue );
               			break;
			 };
		break;

		case 'o':
			if (arglen>=3)
				{
				  	char	* outNamePtr = NULL;
					int	outNameLength = 0;

					outNamePtr=&argv[i][2];
					outNameLength=strlen( outNamePtr );

					if ( strlen( nameOfOutResourceFile )!=0 )
						{
						     puts( "\nError: output font file name already defined. Program stopped.\00" );
  		 				     errorExit( errorExitValue );
						}

	    	  			if ( outNameLength > filePathLength ) 
		     				{
     		 				     puts( "\nError: output font file name is too long. Program stopped.\00" );
  		 				     errorExit( errorExitValue );
		     				};
                  			strncpy( nameOfOutResourceFile, outNamePtr, outNameLength );
					nameOfOutResourceFile[ outNameLength ]=0;
				}
			else
				{
				 	puts( "\nError: output font file name is not defined. Program stopped.\00" );
  		 			errorExit( errorExitValue );
				}
		break;

		case 'i':
			 if (arglen>=3)
				{
				  char	* infoNamePtr = NULL;
				  int	infoNameLength = 0;

				  infoNamePtr=&argv[i][2];
				  infoNameLength=strlen( infoNamePtr );

	    	  		 if ( infoNameLength > filePathLength )
		     			{
     		 			     puts( "\nError: file name of information file is too long. Program stopped.\00" );
  		 			     errorExit( errorExitValue );
		     			};

                  			 strncpy( nameOfInfoResourceFile, infoNamePtr, infoNameLength );
				 nameOfInfoResourceFile[ infoNameLength ]=0;
				}
			else
				{
				 	puts( "\nError: font file name for Unicode information is not defined. Program stopped.\00" );
  		 			errorExit( errorExitValue );
				}
		break;

                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
               };
          }
	else
          {
		puts( "\nError: no flag defined. Program aborted.\00" );
		errorExit( errorExitValue );
          }
     }
     else // Get name of input font file.
     {  char	* inNamePtr = NULL;
	 int	inNameLength = 0;

	 inNamePtr=&argv[i][0];
	 inNameLength=strlen( inNamePtr );

	 if ( strlen( nameOfInResourceFile )!=0 )
		{
		     puts( "\nError: input font file name already defined. Program stopped.\00" );
  		     errorExit( errorExitValue );
		}
	 if ( inNameLength > filePathLength ) 
		{
     		     puts( "\nError: input font file name is too long. Program stopped.\00" );
  		     errorExit( errorExitValue );
		};

               strncpy( nameOfInResourceFile, inNamePtr, inNameLength );
	 nameOfInResourceFile[ inNameLength ]=0;

          // fileSize
      }; // End of if "file" condition;
 }; // End of command line arguments counter;

//puts("");

if ( argc==1 ) 
	usage(); // Display help list;

if ( strlen( nameOfInResourceFile )==0 )
   {
	puts( "\nError: input font file name is not defined. Program stopped.\00" );
	errorExit( errorExitValue );
   }

// Open resource file for read.
resFilePtr = openFile( nameOfInResourceFile, "rb");

(UInt32)fileSize=(UInt32)fileLength( resFilePtr );

if ( resFilePtr==NULL )
   {
     printf("\nError: can not open file '%s'. Program aborted.\n\00", nameOfInResourceFile );
     errorExit( errorFileOperation );
   };

// Check font resource for consistency.
fontVersion = getFontResourceType( resFilePtr, true );
 if ( fontVersion==0 )
	{
	  printf("\nError: file '%s' is not a binary font resource. Program stopped.\n\00", nameOfOutResourceFile );
	  errorExit( errorFileOperation );
	}
else
 if ( fontVersion!=1 )
	{
	  printf("\nError: font resource '%s' is not type version 1. Program aborted.\n\00", nameOfOutResourceFile );
	  errorExit( errorFileOperation );
	};

// get font header.
fileSeek( resFilePtr, 0l ); // Seek file at the start position.
fillAreaByByte(  &fontHeader, 0, sizeof( FontType ) ); // Obtain font header.
getFontHeaderV1WFName( resFilePtr, &fontHeader, nameOfInResourceFile ); // Get font header.
if ( fontHeader.fRectHeight > 127 )
	{
	  printf("\nError: height of glyph exceeds 127 points. Program stopped.\n\00");
  	  errorExit( errorExitValue );
     	};

tablesLength = (fontHeader.lastGlyph-fontHeader.firstGlyph)+3; // Lenght of font tables in words;
(UInt32)fntGlyphBitmapPos=(UInt32)fontTag1Length;
(UInt32)fntBitmapLocationPos=(UInt32)fntGlyphBitmapPos+(UInt32)fontHeader.rowWords*2*(UInt32)fontHeader.fRectHeight;
(UInt32)fntOffsetWidthTblPos=(UInt32)fntBitmapLocationPos+(UInt32)tablesLength*2;

// Examine psf file for allowance to embed.
if (  strlen( nameOfInfoResourceFile )!=0  )
  {
     UInt32 infoFileSize;
     UInt8	  fileFormat=0;

     infoFilePtr = openFile( nameOfInfoResourceFile, "rb" );
     if ( infoFilePtr == NULL )
	{ 
  	  printf("\nError: can not open file '%s'. Program aborted.\n\00", nameOfInfoResourceFile );
  	  errorExit( errorFileOperation );
	};

    (unsigned long)infoFileSize = (unsigned long)fileLength( infoFilePtr ); // Get size of opened file;

     // Check for PSF font format.
    psf1FontHeader.magic[0]=psf2FontHeader.magic[0]=fileGetByteWFName( infoFilePtr , nameOfInfoResourceFile );
    psf1FontHeader.magic[1]=psf2FontHeader.magic[1]=fileGetByteWFName( infoFilePtr , nameOfInfoResourceFile );

    if ( (psf1FontHeader.magic[0]==PSF1_MAGIC0) && (psf1FontHeader.magic[1]==PSF1_MAGIC1) )
	fileFormat=1;
    else
           {
	 psf2FontHeader.magic[2]=fileGetByteWFName( infoFilePtr, nameOfInfoResourceFile );
	 psf2FontHeader.magic[3]=fileGetByteWFName( infoFilePtr , nameOfInfoResourceFile );
	 if ( psf2FontHeader.magic[0]==PSF2_MAGIC0 &&  psf2FontHeader.magic[1]==PSF2_MAGIC1 
		&& psf2FontHeader.magic[2]==PSF2_MAGIC2 && psf2FontHeader.magic[3]==PSF2_MAGIC3 )
	  fileFormat=2;
          };

 // Lets read font header in case magic bytes are for PSF font format.
 switch ( fileFormat )
         {
	  case 1: // Only 1 version supported.
		psf1FontHeader.mode = fileGetByteWFName( infoFilePtr, nameOfInfoResourceFile );	
		psf1FontHeader.charsize = fileGetByteWFName( infoFilePtr, nameOfInfoResourceFile ); // Font Height.
		//(UInt32)infoFileUnicodeDataOffset = (UInt32)sizeof( psfFontV1Type )+(UInt32)psf1FontHeader.charsize*256;

		// Check accordance file length to the calculated size.
		if ( (sizeof(psfFontV1Type)+(psf1FontHeader.charsize*256) )>fileSize )
                              {
		    fileFormat=0;
		    printf("\nError: Incorrect file format. Program aborted.\n\00");
		    fclose( infoFilePtr ); // close input intel x86 raw font data file;
		    errorExit( errorFileOperation );
                              }
		 else
                              {
		   infoFileUnicodeDataOffset = sizeof( psfFontV1Type )+psf1FontHeader.charsize*256;
		   fontHeader.fRectHeight=psf1FontHeader.charsize;
		  }
	  break;

	  case 2: 
		printf("\nError: PSF font version 2 is not supported.\n\00");
		fclose( infoFilePtr ); // close input intel x86 raw font data file;
		errorExit( errorFileOperation );
	  break;
       }
}

// Read font tables;
fileSeek( resFilePtr, (UInt32)fntBitmapLocationPos ); // Seek file at the start position.
for( i=0; i < tablesLength; i++ )
	bitmapLocationTable[ i ] = fileGetWordWFName( resFilePtr, nameOfInResourceFile );

// Obtain offset widthtable.
fileSeek( resFilePtr, (UInt32)fntOffsetWidthTblPos ); // Seek file at the start position.
for( i=0; i < tablesLength; i++ )
      {
	if ( ftell(resFilePtr) < fileSize )
		offsetWidthTable[ i ] = fileGetWordWFName( resFilePtr, nameOfInResourceFile );
	else
		offsetWidthTable[ i ] = 0xFFFF;
      };

tmpFilePtr=tmpfile();
if ( tmpFilePtr==NULL )
	{
          	  printf("\nError: can not open temporary file. Program aborted.\n\00");
	  errorExit( errorFileOperation );
	};

// In case treating PSF file do write its header first from the initial 0 position.
if ( infoFilePtr != NULL )  // Write Magic bytes.
{
 filePutByte( tmpFilePtr, PSF1_MAGIC0 );
 filePutByte( tmpFilePtr, PSF1_MAGIC1 );

 // Write PSF font header version 1.
 filePutByte( tmpFilePtr, psf1FontHeader.mode );
 filePutByte( tmpFilePtr, psf1FontHeader.charsize );
};

// Write raw font data to the temporary file.
for ( i=dosFirstByte; i<=dosLastByte; i++ )
 {
	//  Display warning once when font height is differs from the optional.
	if ( (i==dosFirstByte) && (fontHeader.fRectHeight!=fontHeight) )
	{
	  if ( fontHeader.fRectHeight != fontHeight )
	  	printf( "\n* Warning: the height of font %s than requered for %u pixels.\00", fontHeader.fRectHeight > fontHeight ? "greater" : "less", fontHeader.fRectHeight > fontHeight ? fontHeader.fRectHeight-fontHeight : fontHeader.fRectHeight-fontHeight );
	};

     // When there is no glyph in the source font resource write empty character.
     if ( (i < fontHeader.firstGlyph) || (i > fontHeader.lastGlyph) )
        {
         for( i1=0; i1 < fontHeight; i1++ )
      	     filePutByte( tmpFilePtr, 0x0 );
        }
     else
       {
	 UInt8		dotsToShiftRight=0, dotsToShiftLeft=0, rowCounter; 
	 UInt16		glyphWidth, columnCounter;
	 UInt8		glyphLeftByte, glyphRigthByte;
	 UInt32		filePos;

	 // Create glyph structure.
	 fillAreaByByte( originalGlyphPtr, 0, bufferLength );

	 // Get width from the difference of bitmap locations.
	 originalGlyphPtr->width = bitmapLocationTable[i+1]-bitmapLocationTable[i];

	 if ( originalGlyphPtr->width <= 0 ) // fill character by zero bytes in case glyph is empty.
	   {
	     for(  i1=0; i1 < fontHeight; i1++ )
      	     	filePutByte( tmpFilePtr, 0x0 );
	   }
	 else  // Load glyph from file;
	   {
	     originalGlyphPtr->widthInBytes=(originalGlyphPtr->width+7)>>3;
	     originalGlyphPtr->height=fontHeader.fRectHeight;
	     originalGlyphPtr->bufferPtr=(UInt8 *)originalGlyphPtr+sizeof(glyphBufferType)+1;
	     dotsToShiftLeft=bitmapLocationTable[i] & 7; // How many bits to shift right when add first byte.
	     dotsToShiftRight=(8-dotsToShiftLeft) & 7;   // How many bits to shift left at adding next byte;

	     if ( originalGlyphPtr!=NULL )
	 	for( rowCounter=0; rowCounter < originalGlyphPtr->height; rowCounter++ )
	   		for( columnCounter=0; columnCounter < originalGlyphPtr->widthInBytes; columnCounter++ )
				{
		  			(UInt32)filePos=(UInt32)fntGlyphBitmapPos+(UInt32)(bitmapLocationTable[i]>>3)+(UInt32)columnCounter+(UInt32)rowCounter*(UInt32)fontHeader.rowWords*(UInt32)2;
		  			fileSeek( resFilePtr, (UInt32)filePos );

		  			glyphLeftByte = fileGetByteWFName( resFilePtr, nameOfInResourceFile );
		  			glyphRigthByte = dotsToShiftLeft!=0 ? fileGetByteWFName( resFilePtr, nameOfInResourceFile ) : 0;
		  			originalGlyphPtr->bufferPtr[ (rowCounter * (UInt32)originalGlyphPtr->widthInBytes)+columnCounter ] |=
						( (glyphLeftByte<<dotsToShiftLeft)&0xFF) | ( (glyphRigthByte>>dotsToShiftRight)&0xFF );
				}
	     // Write glyph to the file.
	     for( rowCounter=0; rowCounter < fontHeight; rowCounter++ )
		  if ( rowCounter < originalGlyphPtr->height )
			{
			 if ( originalGlyphPtr->width !=8 )
		 		printf( "\n* Warning: the width of %u glyph %s than requered for %u pixels. Fixed.\00", i, (originalGlyphPtr->width > 8) ? "greater" : "less", (originalGlyphPtr->width > 8) ? (originalGlyphPtr->width-8) : (8-originalGlyphPtr->width) );
			
			 filePutByte( tmpFilePtr, originalGlyphPtr->bufferPtr[ (rowCounter * (UInt32)originalGlyphPtr->widthInBytes) ] & (0xFF << ((originalGlyphPtr->width < 8) ? 8-originalGlyphPtr->width : 0) ) );
			}
	  	  else
      	     		filePutByte( tmpFilePtr, 0x0 );
	   };
         }
 }

// Add Unicode information at the end of file.
if ( infoFilePtr!=NULL )
  {
     // Seek file at the start position and copy all data what goes after original font images.
    fileSeek( infoFilePtr, (UInt32)infoFileUnicodeDataOffset ); 

    while( ftell(infoFilePtr) < fileLength(infoFilePtr) )
	{ 
                 UInt8 unicodeByte;
	   unicodeByte=fileGetByteWFName( infoFilePtr , nameOfInfoResourceFile );
	   filePutByte( tmpFilePtr, unicodeByte );	   
	}
      fclose( infoFilePtr );
  }

// When done.
fclose( resFilePtr ); // Close binary res file;

// In case there is no output file make it.
if ( strlen( nameOfOutResourceFile )==0 )
	{ if ( strlen( nameOfInfoResourceFile )==0 )
		sprintf( nameOfOutResourceFile, "pcfnt8x%u.bin", fontHeight );
	   else
		sprintf( nameOfOutResourceFile, "pcfnt8x%u.psf", fontHeight );
	};

// Move file from temporary file to binary font resource;
fileOutPtr = openFile( nameOfOutResourceFile, "wb+" );

if ( fileOutPtr == NULL )
	{
	  printf("\nError: can not write to file '%s'. Program aborted.\n\00", nameOfOutResourceFile );
	  errorExit( errorFileOperation );
	};

// Copy temporary data to output file.
(unsigned long)fileSize = (unsigned long)fileLength( tmpFilePtr );
printf("\nConvert '%s' to '%s'...", nameOfInResourceFile, nameOfOutResourceFile );
proceedBinaryDataFileTransfer( tmpFilePtr, fileOutPtr, 0l, (unsigned long)fileSize );
printf(" Done.");

fclose( fileOutPtr );
fclose( tmpFilePtr ); // Close temporary res file;

// Pogram is done.
// printf("\nDone.");

 return returnStatus; 
};

// Display user's help.
static void usage( void )
{
  puts("\nUsage ([] means optional):");
  puts("   fnttopc <fontres.bin> [flags]");
  puts("               ... where fontres.bin is a name of font resource to convert.\n");
  puts("Flags are:\n");

  puts("-8            output for 8x8 screen font (2048 bytes file size).");
  puts("-14           output for 8x14 screen font (3584 bytes file size).");
  puts("-16           output for 8x16 screen font (4096 bytes file size) by default.");
  puts("-18           output for 8x18 screen font (4608 bytes file size).");
  puts("-19           output for 8x19 screen font (4864 bytes file size).\n");

  puts("-o<filename>  name of font file for output, by default pcfnt8xHH.(bin|psf).");
  puts("-i<filename>  name of PSF file to import unicode information.\n");
   // puts("NOTICE: The default output file name is pcfnt8xHH, when '-o' option ommited.");
  puts("-h, -?	      this help screen.\n");
  puts("NOTICE: Only one binary font resource can be treated for the program run.");
};
