//  Copyright (C) 2005 Alexander S. Ovcharenko, SymbioForge;
//
//  FntToPdb - this is a command line utility which allows to create a
//  Palm(tm) OS database (.pdb) with font records from binary font resource
//  files.
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//    To compile this file use the following command line options:
//  gcc fnttopdb.c toollib.c -o fnttopdb
//
//    Where GCC is a GNU C compiler or any other suitable to your computer 
//  operation system;

#include <stdio.h>
#include "toollib.h"

#define maxNumberOfResources	32	// Maximum number of resources.

// Local structutre that keeps command line arguments;
typedef struct 
	{
	 char			name[ filePathLength+1 ]; // Name of binary resource file;
	 UInt8			version;
	 UInt8			format;  // resource format;
	 UInt32			length;  // Length of a resource;
	 Boolean		saved;   // When resource already saved.
	} fontResType;

// A local structure with program arguments data.
struct {
	Boolean			helpScreen;
	Boolean			rebuild;
 	char			nameOfPdb[ filePathLength+1 ];
	char			longNameOfPdb[ dbNameLength+1 ];
	fontResType		fontResources[ maxNumberOfResources ];
	short			resourcesNumber; // Number of resource from command line;
	short			originalResourceNumber; // Number of original resources.
	UInt32			lastUniqueID; // Last used unique id.
       } argsData;

// Global variable definition.
pdbHeaderType			dbHeader;
UInt32				listEntryOffset=0, recordEntryOffset=0, nextRecordEntryOffset=0;
pdbRecordEntryType		dbRecordEntry, dbRecordEntryOld;
unsigned long			fileSize, recordSize;
short				resNumber=0;

static void usage( void );	// Display help screen.

main( int argc, char * argv[] )
 { short	i, passNumber, arglen;
   UInt32	uniqID;
   FILE		* tmpFilePtr=NULL, * outFilePtr=NULL, * inFilePtr=NULL, * oldDBFilePtr=NULL;

   puts("FntToPdb - Add binary font resources to Palm OS(tm) database as records." );
   puts("Version 1.3 - (c) 2005, SymbioForge (http://www.symbioforge.com/feii/)." );

   memset( (void *)&argsData, 0, sizeof( argsData ) );
   argsData.originalResourceNumber = argsData.resourcesNumber = 0;

   for( i=1; i < argc; i++ ) // Accomulate arguments.
    {
     arglen=strlen( argv[i] ); // Get length of this argument;
     // Treat command line options;
     if ( ( arglen >= 1 ) && (argv[i][0]=='-' || argv[i][1]=='/') ) // Assume this is a flag;
        { 
          if ( arglen >= 2 ) // Check this flag;
	     switch( argv[ i ][ 1 ] )
               {
		case 'h': // This is help screen request;
		case 'H':
		case '?':
			argsData.helpScreen=1;
		break;

		case 'n': // This is long name of a database;
                case 'N':
			if ( strlen(argsData.longNameOfPdb) > 0 )
			   {
                          	puts( "\nError: the database name already set. Program aborted.\00" );
			  	errorExit( errorExitValue );
			   };

			if ( arglen >=3 ) // Get value;
			   {
			    	char	* dbLongNamePtr=NULL, startQuote, endQuote;
			    	short	dbLongNameLength=0; // Length of database name;

			  	dbLongNamePtr=&argv[i][2];    // Set name of database;
			  	startQuote=dbLongNamePtr[0];  // Get first char of string;
			  	endQuote=dbLongNamePtr[strlen(dbLongNamePtr)-1]; // Get last char of string;

			  	dbLongNameLength=strlen( dbLongNamePtr );

			  	if ( startQuote==endQuote && (startQuote==39 || startQuote==34) && (dbLongNameLength>2) )
			    		{ // Remove quotation marks;
						dbLongNamePtr=&dbLongNamePtr[1];
						dbLongNameLength-=2;
		    	    		};
				if ( dbLongNameLength > dbNameLength )
					{
                         			puts( "\nError: database name is too long. Program stopped.\00" );
			  			errorExit( errorExitValue );
					};

			  	strncpy( argsData.longNameOfPdb, dbLongNamePtr, dbLongNameLength );
			  	argsData.longNameOfPdb[ dbLongNameLength ] = 0;
			    }
                          else
			    {
                         	puts( "\nError: database name is not defined. Program stopped.\00" );
			  	errorExit( errorExitValue );
			    };
                break;

                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
              } // End of switch block;
	    else  // End of argument length block;
	     {
		puts( "\nError: incorrect use of argument. Program aborted.\00" );
		errorExit( errorExitValue );
	     };
        } // End of flag checking;
      else
      if ( ( arglen >= 1 ) && (argv[i][0]=='+') ) // In case the 1st char is '+' assume that font binarie follows;
	{ char * fontBinaryNamePtr = NULL;

	  fontBinaryNamePtr=&argv[i][1];

	  if ( strlen( fontBinaryNamePtr ) > filePathLength )
		{
		 puts ("\nError: file name of font binary resource is long. Program aborted.\00" );
		 errorExit( errorExitValue );
		}
          else // Check font binary resource, get resource information;
		{ FILE * resFilePtr=NULL;

		  // Open file defined in the command line arguments.
		  resFilePtr = openFile( fontBinaryNamePtr, "rb" );

		  if ( resFilePtr == NULL )
			{
		  	  printf("\nError: can not open file '%s'. Program aborted.\n\00", fontBinaryNamePtr );
		  	  errorExit( errorFileOperation ); 
			};

	  	  // Obtain version of font resource; Store resource information;
	  	  if ( (argsData.fontResources[ argsData.resourcesNumber ].version = getFontResourceType( resFilePtr, true ) )!=0 )
	        	{
		 	  if ( argsData.resourcesNumber >= maxNumberOfResources )
				{ 
				 puts("\nError: maximum allowed number of binary resources is reached. Program stopped.\00");
				 errorExit( errorExitValue );
				};

		 	   // Store name of binary resource file to array.
		 	   strcpy( argsData.fontResources[ argsData.resourcesNumber ].name, fontBinaryNamePtr );
		 	   argsData.fontResources[ argsData.resourcesNumber ].length = fileLength( resFilePtr ); // Size of binary resource.
		 	   argsData.resourcesNumber++;
	        	}
 	  	  else
	        	{
		  	  printf("\nError: file '%s' is not a binary font resource. Program aborted.\n\00", fontBinaryNamePtr );
		  	  errorExit( errorFileOperation ); 
			};

		  fclose( resFilePtr ); // Close checked resource.
		 }; // End of checking font binary resource.
	  }
       else // Proceed file treat.
          {
	   if ( strlen ( argsData.nameOfPdb ) > 0 ) 
		{
		 puts ("\nError: file name of database already defined. Program stopped.\00" );
		 errorExit( errorExitValue );
		}
           else
	   if ( arglen > filePathLength )
		{
		 puts ("\nError: name of database file is too long. Program aborted.\00" );
		 errorExit( errorExitValue );
		};
	   strncpy( argsData.nameOfPdb, argv[ i ], arglen );
	  };
   }; // End of for cycle. Accomulating arguments.

if ( strlen( argsData.nameOfPdb ) == 0 && argsData.resourcesNumber == 0 )
  usage();
else
if ( argsData.resourcesNumber == 0 )  
   {
     puts( "\nError: no binary font resources defined. Program aborted.\00" );
     errorExit( errorExitValue );
   }
else
   {
     fillAreaByByte( (void *)&dbHeader, 0, pdbHeaderLength );
     tmpFilePtr = tmpfile(); // Create temporary file.
     if (tmpFilePtr==NULL)
	{
	  printf("\nError: can not open temporary file. Program aborted.\n\00");
	  errorExit( errorFileOperation );
	};

     // Try to read data from previous font database.
     oldDBFilePtr = fopen( argsData.nameOfPdb, "rb" );
     if ( oldDBFilePtr!=NULL ) // Treat it if file exists.
	{
	 argsData.rebuild = false;
	 (unsigned long)fileSize=(unsigned long)fileLength( oldDBFilePtr );
	 fileGetPdbHeader( &dbHeader, oldDBFilePtr );
	 if ( dbHeader.recordList.nextRecordList!=0 )
		{
	  	  printf("\nError: can not support multiply record lists. Program stopped.\n\00");
	  	  errorExit( errorFileOperation );
		};

	 if ( strlen(argsData.longNameOfPdb) !=0 )
 	 	strcpy( dbHeader.dbName, argsData.longNameOfPdb );

	 if ( (UInt32)dbHeader.dbType != (UInt32)fontDatabaseID0 &&
	       (UInt32)dbHeader.dbType != (UInt32)fontDatabaseID1 &&
	       (UInt32)dbHeader.dbType != (UInt32)fontDatabaseID2 )
		{
	   	  printf("\nError: file '%s' is not a font database. Program aborted.\n\00", argsData.nameOfPdb );
	   	  errorExit( errorFileOperation );
		};

	  // Change database header values;
 	  if ( ( (UInt32)dbHeader.modifiedTime!=(UInt32)timePalm() ) && ( (UInt32)dbHeader.modifiedTime!=0 ) )
 	        (UInt32)dbHeader.modifiedTime = (UInt32)timePalm();

	  dbHeader.modNumber++;
	  argsData.originalResourceNumber = dbHeader.recordList.numRecords;
	  dbHeader.recordList.numRecords += argsData.resourcesNumber;
          filePutPdbHeader( tmpFilePtr, &dbHeader ); // Write pdb header;
	}
      else
	{
      	  // Build database header;
 	  strcpy( dbHeader.dbName, strlen( argsData.longNameOfPdb )!=0 ? argsData.longNameOfPdb : argsData.nameOfPdb );
 	  (UInt32)dbHeader.createTime = (UInt32)timePalm();
	  (UInt32)dbHeader.id = (UInt32)dbHeader.dbType = (UInt32)fontDatabaseID1; // Set database type;
	  dbHeader.modNumber = 1;
	  dbHeader.version = 1;
          dbHeader.recordList.numRecords = argsData.resourcesNumber;
          filePutPdbHeader( tmpFilePtr, &dbHeader ); // Write pdb header;
     	  argsData.rebuild = true;
	};

	// Use short database name for long name if it's empty.
	if ( strlen(argsData.longNameOfPdb) ==0 )
		strcpy( argsData.longNameOfPdb, argsData.nameOfPdb ); // Write font database;

        // Treat database.
        printf("\nFilename of database to %s: '%s'\n", argsData.rebuild ? "build": "add to", argsData.nameOfPdb );
 	printf("Complete database name for Palm OS: '%s'\n", argsData.longNameOfPdb );

        for( passNumber=0; passNumber < 2; passNumber++ )
         {
          dbRecordEntry.recordOffset = pdbHeaderLength+sizeof( pdbRecordEntryType )*dbHeader.recordList.numRecords; // Reduce a word (firstEntry field size);
          for( i=0, uniqID=0x128001; i < dbHeader.recordList.numRecords; i++, uniqID++ )
	    { 
	     if ( passNumber==0 ) // Write recordEntries for the 1'st pass.
	      {
		// Retreive records from old database.
		if ( ( i < argsData.originalResourceNumber ) && ( oldDBFilePtr!=NULL ) ) 
		   { 
			listEntryOffset=fileGetPDBListEntryOffset( 0, oldDBFilePtr ); //
			(UInt32)recordEntryOffset = (UInt32)fileGetPDBRecordOffset( listEntryOffset, i, oldDBFilePtr ); //
			// Get record entry information;
			fileSeek( oldDBFilePtr, recordEntryOffset+2 );
			dbRecordEntry.attributes=fileGetByte( oldDBFilePtr );
			dbRecordEntry.uniqueID[0]=fileGetByte( oldDBFilePtr );
			dbRecordEntry.uniqueID[1]=fileGetByte( oldDBFilePtr );
			dbRecordEntry.uniqueID[2]=fileGetByte( oldDBFilePtr );
			(UInt32)uniqID = dbRecordEntry.uniqueID[0]*0x10000;
			(UInt32)uniqID += dbRecordEntry.uniqueID[1]*0x100;
			(UInt32)uniqID += dbRecordEntry.uniqueID[2];

			(UInt32)nextRecordEntryOffset = (UInt32)fileGetPDBRecordOffset( listEntryOffset, i+1, oldDBFilePtr ); //
			(UInt32)recordSize = ((UInt32)nextRecordEntryOffset == 0l ? (UInt32)fileSize : (UInt32)nextRecordEntryOffset )-(UInt32)recordEntryOffset;

	  		filePutRecordEntry( &dbRecordEntry, i, tmpFilePtr );
			(UInt32)dbRecordEntry.recordOffset+=(UInt32)recordSize;
		   }
		else
		   {
			resNumber = i - argsData.originalResourceNumber;
	  		dbRecordEntry.uniqueID[ 0 ] = (UInt8)( 0xFF&(uniqID/0x10000) );
	  		dbRecordEntry.uniqueID[ 1 ] = (UInt8)( 0xFF&(uniqID/0x100) );
	  		dbRecordEntry.uniqueID[ 2 ] = (UInt8)( 0xFF&uniqID );
	  		dbRecordEntry.attributes = 0x40;
	  		filePutRecordEntry( &dbRecordEntry, i, tmpFilePtr );
			(UInt32)dbRecordEntry.recordOffset+=(UInt32)argsData.fontResources[ resNumber ].length;
		   };
	      }
	     else
	     if ( passNumber==1 )
	      {
		if ( ( i < argsData.originalResourceNumber ) && ( oldDBFilePtr!=NULL ) )
		  {
			listEntryOffset = fileGetPDBListEntryOffset( 0, oldDBFilePtr ); //
			(UInt32)recordEntryOffset = (UInt32)fileGetPDBRecordOffset( listEntryOffset, i, oldDBFilePtr ); //
			(UInt32)nextRecordEntryOffset = (UInt32)fileGetPDBRecordOffset( listEntryOffset, i+1, oldDBFilePtr ); //
			(UInt32)recordSize = ((UInt32)nextRecordEntryOffset == 0l ? (UInt32)fileSize : (UInt32)nextRecordEntryOffset )-(UInt32)recordEntryOffset;
			proceedBinaryDataFileTransfer( oldDBFilePtr, tmpFilePtr, (UInt32)recordEntryOffset, (UInt32)recordSize );
		  }
		else
		  {
			resNumber = i - argsData.originalResourceNumber;
			printf("%s binary font resource: '%s', ", oldDBFilePtr==NULL? "Link":"Add", argsData.fontResources[ resNumber ].name );
			switch ( argsData.fontResources[ resNumber ].version )
			 {
			   case type1NFNT:
				puts("NFNT (v1).");
			   break;
			   case type1AFNX:
				puts("AFNX (v1).");
			   break;
			   case type2nfnt:
				puts("nfnt (v2).");
			   break;
			   case type2afnx:
				puts("afnx (v2).");
			   break;
			 };

			// Open external resource.
			inFilePtr = NULL;
			inFilePtr = openFile( (char *)argsData.fontResources[ resNumber ].name, "rb" );
			if ( inFilePtr == NULL )
				{
	  		  	 printf("\nError: can not open file '%s'. Program aborted.\n\00", argsData.fontResources[ resNumber ].name );
	 		  	 errorExit( errorFileOperation );
				};
			proceedBinaryDataFileTransfer( inFilePtr, tmpFilePtr, 0l, (unsigned long)argsData.fontResources[ resNumber ].length );
			fclose( inFilePtr ); // Close binary resource file.

			argsData.fontResources[ resNumber ].saved=1;
	      	  }; 
	 	};
	      }; // End of record counter;

	   if ( passNumber==0 ) // Add a word at the end of entry list;
	    {
	      fileSeek( tmpFilePtr, ((UInt32)sizeof(pdbHeaderType)+(UInt32)sizeof(pdbRecordEntryType)*dbHeader.recordList.numRecords)-sizeof(UInt16) );
    	      filePutWord( tmpFilePtr, 0 );
	    };
        }; // End of pass counter;

	 // Close old database file opened for read.
	if ( oldDBFilePtr!=NULL )
		fclose( oldDBFilePtr );

	// Move file from temporary data file to database.
	// Open pdb file for write with truncate.
	outFilePtr = openFile( argsData.nameOfPdb, "wb+" );
	if ( outFilePtr==NULL )
	  {
	    printf("\nError: can not open file '%s' for write. Program aborted\n\00", argsData.nameOfPdb );
	    errorExit( errorFileOperation );
	  };

	(unsigned long)fileSize=(unsigned long)fileLength( tmpFilePtr );
	proceedBinaryDataFileTransfer( tmpFilePtr, outFilePtr, 0l, (unsigned long)fileSize );

        fclose( tmpFilePtr ); // Close temporary work file.
	fclose( outFilePtr ); // Close font database.

   }; //End of condition;

  if ( argsData.helpScreen ) usage();
  return;
};

// Display a usage help screen.
static void usage( void ) // Display help screen.
 {
  puts("\nUsage ([] means optional):");
  puts("   fnttopdb [flags] <file.pdb> <+fontres.bin>...");
  puts("                ...where <file.pdb> is a Palm OS(tm) database file to build.");
  puts("                         <fontres.bin> is a raw binary font resource to add.\n");
  puts("Flags are:\n");
  puts("-n<name>   Full Palm OS(tm) database name w/o spaces.");
  puts("-h, -?	   Help screen.");
  return;
 };
