//  Copyright (C) 2004 Alexander S. Ovcharenko, SymbioForge;
//
//  Fnt2Prc - utility which allows to add binary font resource to .prc files.
//
//  This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
//  gcc fnttoprc.c toollib.c -o fnttoprc
//
//   Where GCC is a GNU C compiler or other used by your coumputer operation
//  system;

#include <stdio.h>
#include "toollib.h"

#define maxNumberOfResources	32

typedef struct 
	{
	 char			name[ filePathLength+1 ];
	 UInt8			version;
	 UInt32			resourceType;
	 UInt16			resourceID;
	 UInt32			length;  // Length of a resource;
	 Boolean		saved;   // When resource already saved.
	} fontResourceType;

// A structure with program arguments data.
struct {
	Boolean			helpScreen;
	Boolean			createNewResDBFile;
	char			nameOfResDBFile[ filePathLength+1 ];
	char			longNameOfPdb[ dbNameLength+1 ];
	fontResourceType	fontResources[ maxNumberOfResources ];
	UInt32			databaseCreatorID;
	UInt16			resourcesNumber;
	UInt16			resourcesToReplace;
	UInt16			resourcesToAdd;
	UInt16			oldResourcesNumber;
       } argsData;

pdbHeaderType			dbHeader;
FILE 	 			* inFilePtr=NULL, *outFilePtr=NULL, *tmpFilePtr=NULL;
FILE 				* fileBinResPtr=NULL;
unsigned long			fileSize=0;
unsigned long			resourceSize=0;
unsigned long			listEntryOffset=0;
unsigned long			fileDataPos=0;
resourceEntryType		resourceEntry, nextResourceEntry;

static void usage( void );	// Display help screen.
static void indicateResourceLink( UInt32, UInt16, Boolean );

main( int argc, char * argv[] )
 { int		i, arglen;
   UInt16	overrideValue = 0, overrideFontType = 0, resCounter;
   Boolean	nextFilesAreResources = 0;

   puts("FNTTOPRC - Add binary font resources to a Palm OS(tm) resource database." );
   puts("Version 1.3 - (c) 2005, SymbioForge (http://www.symbioforge.com/feii/)." );

   // Initialize arguments data.
   // memset( (void *)&argsData, 0, sizeof( argsData ) );
   fillAreaByByte( &argsData, 0, sizeof( argsData ) );
   argsData.resourcesNumber = 0;
   argsData.createNewResDBFile = false;

   for( i=1; i < argc; i++ )
    {
     // Treat command line options;
     arglen = strlen(argv[i]);

     if ( ( arglen >= 1 ) && ( argv[i][0]=='-' || argv[i][0]=='/' ) )// Assume this is a flag;
        { 
          if ( arglen >= 2 ) // Check the flag;
	    {
	     switch( argv[ i ][ 1 ] )
               {
		case '?':
		case 'h':
		case 'H':
			argsData.helpScreen=1;
			// usage();
		break;

		case 'R':
                case 'r':
			if ( arglen >=3 ) // Get value;
			 {
			  UInt16 value=0;
			  value=atoi( &argv[ i ][ 2 ] );
			  if ( value < 1000 )
				{
                             	 puts( "\nError: override resource number is less than 1000. Program stopped.\00" );
				 errorExit( errorExitValue );
				}
			  else
				overrideValue = value;
		         }
                        else
			 {
                          puts( "\nError: resource start value is not set. Program stopped.\00" );
			  errorExit( errorExitValue );
			 };
                break;

		case 'c':
		case 'C':
			if ( arglen >= (2+4) ) //Get value;
			 {
			   if (argsData.databaseCreatorID!=0)
			    {
                             puts( "\nError: database creator ID already defined. Program aborted.\00" );
			     errorExit( errorExitValue );
			    }
			   else
			      (UInt32)argsData.databaseCreatorID=(UInt32)((UInt32)argv[i][2]<<24)+(UInt32)((UInt32)argv[i][3]<<16)+(UInt32)((UInt32)argv[i][4]<<8)+(UInt32)(argv[i][5]);
		         }
			else
			 {
                            puts( "\nError: incorrect database creator ID. Program stopped.\00" );
			    errorExit( errorExitValue );
			 };
		break;

		case 't':
		case 'T':
		case 'a':
		case 'A':
			if ( arglen >= 3  ) //Get value;
			 { UInt16 value=0;
			   value = atoi( &argv[ i ][ 2 ] );
			   if ( value==1 || value==2 )
			      {
				 if ( (argv[i][1]=='T') || (argv[i][1]=='t') )
				   {
					if ( value==1 )
			     	 	  overrideFontType = type1NFNT;
					else
					  overrideFontType = type2nfnt;
				   };
				 if ( (argv[i][1]=='A') || (argv[i][1]=='a') )
				   {
					if ( value==1 )
					  overrideFontType = type1AFNX;
					else
					  overrideFontType = type2afnx;
				   };
			      }
			   else
			      {
                             	  puts( "\nError: font type value is invalid. Program stopped.\00" );
			     	  errorExit( errorExitValue );
			      };
			 }
			else
			 {
                             puts( "\nError: font type value is not defined. Program stopped.\00" );
			     errorExit( errorExitValue );
			 };
		break;

		case 'n': // This is long name of a database;
                case 'N':
			if ( strlen(argsData.longNameOfPdb) > 0 )
			   {
                          	puts( "\nError: the database name already set. Program aborted.\00" );
			  	errorExit( errorExitValue );
			   };

			if ( arglen >=3 ) // Get value;
			   {
			    	char	* dbLongNamePtr=NULL, startQuote, endQuote;
			    	short	dbLongNameLength=0; // Length of database name;

			  	dbLongNamePtr=&argv[i][2];    // Set name of database;
			  	startQuote=dbLongNamePtr[0];  // Get first char of string;
			  	endQuote=dbLongNamePtr[strlen(dbLongNamePtr)-1]; // Get last char of string;

			  	dbLongNameLength=strlen( dbLongNamePtr );

			  	if ( startQuote==endQuote && (startQuote==39 || startQuote==34) && (dbLongNameLength>2) )
			    		{ // Remove quotation marks;
						dbLongNamePtr=&dbLongNamePtr[1];
						dbLongNameLength-=2;
		    	    		};
				if ( dbLongNameLength > dbNameLength )
					{
                         			puts( "\nError: database name is too long. Program stopped.\00" );
			  			errorExit( errorExitValue );
					};

			  	strncpy( argsData.longNameOfPdb, dbLongNamePtr, dbLongNameLength );
			  	argsData.longNameOfPdb[ dbLongNameLength ] = 0;
			    }
                          else
			    {
                         	puts( "\nError: database name is not defined. Program stopped.\00" );
			  	errorExit( errorExitValue );
			    };
                break;

                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
               }; // End of switch block.
          } // End of argument length case;
	else
	 {
		puts( "\nError: incorrect use of argument. Program aborted.\00" );
		errorExit( errorExitValue );
	 };
      }
   else // Assume this is a resource adding flag;
   if ( (arglen>=1) && (argv[i][0]=='+') ) 
      { // Proceed binary font resource treat.
	UInt16	resourceID=0;
	UInt32	resourceType=0;
	char *	hexValueStrPtr=NULL;

	if ( strlen( &argv[i][1] ) >= filePathLength )
	   {
		printf("\nError: filename length of resource exceeds %u characters. Program aborted.\n\00", filePathLength);
		errorExit( errorFileOperation );
	   };
	if ( argsData.resourcesNumber >= maxNumberOfResources )
	   { 
		printf("\nError: maximum allowed number of binary resources reached. Program stopped.\n\00");
		errorExit( errorExitValue );
  	   };

	strcpy( argsData.fontResources[ argsData.resourcesNumber ].name, &argv[i][1] );
	// Open the resource file defined in the command line arguments.
	fileBinResPtr = openFile( &argv[i][1], "rb" );
	argsData.fontResources[ argsData.resourcesNumber ].length = fileLength( fileBinResPtr ); // Get size of treated file;
	argsData.fontResources[ argsData.resourcesNumber ].version = getFontResourceType( fileBinResPtr, true );
	fclose( fileBinResPtr ); // Close checked file.

	if ( argsData.fontResources[ argsData.resourcesNumber ].version == 0 )
	   {
	    printf( "\nError: file '%s' is not a binary font resource. Program aborted.\n", &argv[i][1] );
	    errorExit( errorFileOperation );
	   };

	// Retreive resource number from filename.;
	if ( strlen(argsData.fontResources[ argsData.resourcesNumber ].name) > 4 )
	   {
	    hexValueStrPtr = &argsData.fontResources[ argsData.resourcesNumber ].name[4];
	    if ( strlen(hexValueStrPtr)!=0 ) resourceID=(UInt16)HtoI( hexValueStrPtr );
	   }
	else
	   resourceID=1000;

	argsData.fontResources[ argsData.resourcesNumber ].resourceID = ( overrideValue ==0 ) ? resourceID : overrideValue;
	argsData.fontResources[ argsData.resourcesNumber ].saved = false; // Resource is not saved yet.

	if ( overrideFontType!=0 ) // Override font type;
		argsData.fontResources[ argsData.resourcesNumber ].version = overrideFontType;

	(UInt32)argsData.fontResources[ argsData.resourcesNumber ].resourceType = (UInt32)fontVersionToResType( argsData.fontResources[ argsData.resourcesNumber ].version );

	argsData.resourcesNumber++;
	overrideValue = overrideFontType = 0; // Initialize override values.
      }
   else // Proceed PalmOS database treat.
      {
	if ( arglen >= filePathLength )
	   {
		printf("\nError: filename length of database exceeds %u characters. Program aborted.\n\00", filePathLength );
		errorExit( errorFileOperation );
	   };
   	if ( strlen( argsData.nameOfResDBFile )!=0 )
	   {
 		printf("\nError: resource database file already defined. Program stopped.\n\00");
		errorExit( errorExitValue );
	   };
	strcpy( argsData.nameOfResDBFile, argv[ i ] );

	// Open file defined in the command line arguments.
	argsData.createNewResDBFile = false;
	inFilePtr = fopen( argv[i], "rb" );
	if ( inFilePtr == NULL )
	   {
		  argsData.createNewResDBFile = true;
		  continue; // continue to iterate next command line arguments.
	   };

	 fileSize = fileLength( inFilePtr ); // Get size of treated file;
	 if ( fileSize < pdbHeaderLength )
	   {
		  printf("\nError: can not proceed the file '%s'. Program aborted.\n\00", argv[i]); 
		  errorExit( errorFileOperation ); 
	   };

	 // Check header for type of databse.
	 fillAreaByByte( &dbHeader, 0, lengthOfpdbHeader );
	 fileGetPdbHeader( &dbHeader, inFilePtr );

	 (unsigned long)listEntryOffset = (unsigned long)fileGetPDBListEntryOffset( 0, inFilePtr );

	 // Get format of database ( records or resources );
	 // Check for mutiple record lists. It is unusal if it have offset on the next recordlist.
	 if ( ( fileGetPDBResourceFormat( listEntryOffset, inFilePtr )!=resourcesDataFormat ) || ( fileGetPDBNextRecordListOffset( (unsigned long)listEntryOffset, inFilePtr )!=0l ) )
	     {
	    	printf("\nError: unsupported format of '%s' font database. Program stopped.\n", argv[i] );
	    	errorExit( errorExitValue );
	     };
	 fclose( inFilePtr ); // Close checked file.

  	}; // End of else - file treat;
  }; // End of for(x;x;x) line...

if ( argc==1 ) usage(); // Display help screen if no arguments defined.
   else
if ( strlen( argsData.nameOfResDBFile )==0 )
     {
	  printf("\nError: no resource database file defined. Program stopped.\n\00");
	  errorExit( errorExitValue );
     }
   else
if ( argsData.resourcesNumber<=0 )
     {
	   printf("\nError: no font binary resource defined. Program stopped.\n\00");
	   errorExit( errorExitValue );
     }
else
     { 
	   UInt16 i, i0, i1;
	   printf("\nFile name of database to %s: %s\n", argsData.createNewResDBFile ? "build" : "add", argsData.nameOfResDBFile );
	   printf("Resource database name for PalmOS: %s\n", strlen( argsData.longNameOfPdb )!=0 ? argsData.longNameOfPdb : argsData.nameOfResDBFile );

	   // Open temporary file for write operations;
	   tmpFilePtr = tmpfile(); // open temporary file;
	   if ( tmpFilePtr == NULL )
		{
		  printf("\nError: can not create temporary file. Program aborted.\n\00");
		  errorExit( errorFileOperation );
		};

	   // Open prc file for read.
	   inFilePtr = fopen( argsData.nameOfResDBFile, "rb" );

	   fillAreaByByte( &dbHeader, 0, pdbHeaderLength );
	   argsData.resourcesToReplace=0;

	   if ( inFilePtr == NULL ) // In case file can not be opened.
		{
		  // Create database header.
		  strcpy( dbHeader.dbName, strlen( argsData.longNameOfPdb )!=0 ? argsData.longNameOfPdb : argsData.nameOfResDBFile );
 	  	  (UInt32)dbHeader.createTime = (UInt32)timePalm();

		  dbHeader.flags=1; // Set flag - PRC Executable.
		  dbHeader.version=1;
		  (UInt32)dbHeader.dbType = (UInt32)fontDatabaseID1; // Set database type;
	  	  (UInt32)dbHeader.id = (UInt32)fontsForOS5;
	  	  dbHeader.modNumber = 1;
		}
	   else
		{
	   	  fileGetPdbHeader( &dbHeader, inFilePtr );

		  // Computate number of resources to replace.
	          for ( i=0; i < dbHeader.recordList.numRecords; i++ )
	    	      {
		         fileGetResourceEntry( inFilePtr, &resourceEntry, i );
		         for( i0=0; i0 < argsData.resourcesNumber; i0++ )
		            {
			      if ( resourceEntry.id == argsData.fontResources[ i0 ].resourceID
				   && (UInt32)resourceEntry.type == (UInt32)argsData.fontResources[ i0 ].resourceType )
			          	{
				   	 argsData.resourcesToReplace++;
			           	 break;
			          	};
		            };
	               };

		  if ( strlen( argsData.longNameOfPdb )!=0 )
		  	strcpy( dbHeader.dbName, argsData.longNameOfPdb );
		};

	   if   (argsData.databaseCreatorID !=0 )
	  	  (UInt32)dbHeader.id = (UInt32)argsData.databaseCreatorID;
	   if   (argsData.databaseCreatorID !=0 ) 
		  printf("Database creator ID set to: '%c%c%c%c'\n", (char)((UInt32)dbHeader.id>>24)&0xFF, (char)((UInt32)dbHeader.id>>16)&0xFF, (char)((UInt32)dbHeader.id>>8)&0xFF, (char)((UInt32)dbHeader.id)&0xFF );

	   argsData.resourcesToAdd = argsData.resourcesNumber-argsData.resourcesToReplace;
	   argsData.oldResourcesNumber = dbHeader.recordList.numRecords;

	   // Correct database header for number of records.
           dbHeader.recordList.numRecords += argsData.resourcesToAdd;
	   filePutPdbHeader( tmpFilePtr, &dbHeader );


           // Begin of multipass write to temporary prc;
           //----------------------------------------------------------------------------
	   // Double pass to write entry and data to a work .prc file;
	   for( i=0; i < 2; i++ )
	      {
   	        // Get offset of file for the 1'st resource.
	        fileDataPos = (lengthOfpdbHeader)+dbHeader.recordList.numRecords * resourceEntryLength;

	   	// Treat all resources in old .prc file and find what resource to replace.
	   	i0=0; // Step by step read and overwrite resources.
	   	while( i0 < argsData.oldResourcesNumber )
		    {
			// Get information on resources in old database.
		  	fileGetResourceEntry( inFilePtr, &resourceEntry, i0 ); // Once more find entry to replace;
		  	for( i1=0; i1<argsData.resourcesNumber; i1++ )
			  {
		      	    if ( resourceEntry.id == argsData.fontResources[ i1 ].resourceID
			   	&& (UInt32)resourceEntry.type == (UInt32)argsData.fontResources[ i1 ].resourceType )
				 {
				   // resourceEntry.id = argsData.fontResources[ i1 ].resourceID;
				   // (UInt32)resourceEntry.type = (UInt32)argsData.fontResources[ i1 ].resourceType;
			           break; // We have the same resource in the arguments, so replace it.
				 };
			  };

		  	// Calculate size of a resource;
		  	if ( i1 < argsData.resourcesNumber ) // Replace the resource.
	       			(unsigned long)resourceSize = (unsigned long)argsData.fontResources[ i1 ].length;
		  	else
			 { 	// Get size from the prevous database resource.
		   	 	if ( (i0+1) < argsData.oldResourcesNumber )
					{	
		       				fileGetResourceEntry( inFilePtr, &nextResourceEntry, i0+1 );
					        (UInt32)resourceSize=(UInt32)nextResourceEntry.offset-(UInt32)resourceEntry.offset;
						// When resource size is greater than file size display a message.
						if ( (UInt32)resourceSize > (UInt32)fileSize )
							{
								printf("\nError: invalid source PalmOS database. Program stopped.\n\00");
		  						errorExit( errorFileOperation );
							};
					}
				else
		      			(unsigned long)resourceSize=(unsigned long)fileSize-(unsigned long)resourceEntry.offset;
			 };

			if ( i==0 )
                          {
  	  	     	    (unsigned long)resourceEntry.offset = (unsigned long)fileDataPos; // Position of file data.
		  	    filePutResourceEntry( &resourceEntry, i0, tmpFilePtr );
 			  }
			else
			   {
			      // Here we have got all information on resource for update.
		  	      // Use old resource or replace by new resource.
		  	      if ( i1 < argsData.resourcesNumber ) // Replace the resource.
			      	{ 
				 fileBinResPtr = openFile( (char *)argsData.fontResources[ i1 ].name, "rb" );
			         if ( fileBinResPtr == NULL )
			   	   {
		  		    printf("\nError: can not open file '%s'. Program aborted.\n\00", argsData.fontResources[ i1 ].name );
		 		    errorExit( errorFileOperation );
			   	   };
		  	         fileSeek( tmpFilePtr, fileDataPos );
	 		         proceedBinaryDataFileTransfer( fileBinResPtr, tmpFilePtr, 0l, (unsigned long)resourceSize );
			         fclose( fileBinResPtr ); // Close binary resource file.
			         argsData.fontResources[ i1 ].saved=true;
			         indicateResourceLink( (UInt32)argsData.fontResources[ i1 ].resourceType, argsData.fontResources[ i1 ].resourceID, false );
		                }
                 	      else
		    	        {   // Write old resource to working file.
		  	             fileSeek( tmpFilePtr, fileDataPos );
			             proceedBinaryDataFileTransfer( inFilePtr, tmpFilePtr, resourceEntry.offset, (unsigned long)resourceSize );
		    	        };
		   	  };

		         (unsigned long)fileDataPos += (unsigned long)resourceSize;
		         i0++;
		    }; // End of while (i0) loop;

		 // Add resource what left after replacements.
		 i1=0;
	 	 for( i0=0; i0 < argsData.resourcesNumber; i0++ )
		   if ( !argsData.fontResources[ i0 ].saved )
   		     {
			(unsigned long)resourceSize = (unsigned long)argsData.fontResources[ i0 ].length;
			if ( i==0 )
			  {
	  		        (unsigned long)resourceEntry.offset = (unsigned long)fileDataPos; // Position of file data.
				(UInt16)resourceEntry.id = (UInt16)argsData.fontResources[ i0 ].resourceID;
				(UInt32)resourceEntry.type = (UInt32)argsData.fontResources[ i0 ].resourceType;
				filePutResourceEntry( &resourceEntry, argsData.oldResourcesNumber+i1, tmpFilePtr );
				i1++;
			  }
			else
			if ( i==1 )
			  {
				// Open external resource.
				fileBinResPtr = openFile( (char *)argsData.fontResources[ i0 ].name, "rb" );
				if ( fileBinResPtr == NULL )
		   	   		{
	  					printf("\nError: can not open file '%s'. Program aborted.\n\00", argsData.fontResources[ i0 ].name );
	 					errorExit( errorFileOperation );
		   	   		};
		  	        fileSeek( tmpFilePtr, fileDataPos );
 				proceedBinaryDataFileTransfer( fileBinResPtr, tmpFilePtr, 0l, (unsigned long)resourceSize );
				fclose( fileBinResPtr ); // Close binary resource file.
				argsData.fontResources[ i0 ].saved=true; 
			        indicateResourceLink( (UInt32)argsData.fontResources[ i0 ].resourceType, argsData.fontResources[ i0 ].resourceID, true );
			  };
		      fileDataPos += resourceSize;
   		     }; // End of conditions.
           }; // End of multipass block.

    // Close old prc file if it opened.
    if ( inFilePtr != NULL ) fclose( inFilePtr );

    // Store time of database modification;
    if ( dbHeader.modifiedTime!=timePalm() && dbHeader.modifiedTime!=0 )
	dbHeader.modifiedTime=timePalm();

    filePutPdbHeader( tmpFilePtr, &dbHeader );

    // Move file from data file to prc; Open prc file for write with truncate.
    outFilePtr = openFile( argsData.nameOfResDBFile, "wb+" );
    if ( outFilePtr == NULL )
	{
	  printf("\nError: can not open file '%s' for write. Program aborted.\n\00", argsData.nameOfResDBFile );
	  errorExit( errorFileOperation );
	};
    (unsigned long)fileSize = (unsigned long)fileLength( tmpFilePtr ); // Get size of temporary file;
    proceedBinaryDataFileTransfer( tmpFilePtr, outFilePtr, 0l, (unsigned long)fileSize );

    fclose( outFilePtr ); // Close temporary work file.
   };

 if ( argsData.helpScreen ) usage(); // Display help screen if requested.
    return;
};

// Display a usage help screen.
static void usage( void ) // Display help screen.
 {
  puts("\nUsage ([] means optional):");
  puts("   fnttoprc <file.prc> [-h] <[flags] +fontres.bin>...");
  puts("                  ...where <file.prc> is a Palm OS resources database file.");
  puts("                           <fontres.bin> is a binary font resources.\n");
  puts("Flags are:\n");
  puts("-h, -?	    Help screen.");
  puts("-n<name>    Full PalmOS(tm) name of the database w/o spaces.");
  puts("-c<value>   A four characters value of database creator ID (default: FOS5).");
  puts("-r<value>   Override number of resource for the following binary file only.");
  puts("-t|a<value> Override resource type: -t1 is type 1 (NFNT), -t2 type 2 (nfnt).");
  puts("                                    -a1 is type 1 (AFNX), -a2 type 2 (afnx).");
  return;
 };

// Indicates resource linking.
static void indicateResourceLink( UInt32 resourceType, UInt16 resourceID, Boolean isResourceAdded )
{
 printf("\nBinary font resource (" );
 switch( resourceType )
  {
	case 'NFNT':
		printf("NFNT");
	break;

	case 'nfnt':
		printf("nfnt");
	break;

	case 'AFNX':
		printf("AFNX");
	break;

	case 'afnx':
		printf("afnx");
	break;

	default:
		printf("unknown");
	break;
  };
 printf(" , ID %u) - %s.", resourceID, isResourceAdded ? "added" : "replaced" );
return;
};