//  Copyright (C) 2005 Alexander S. Ovcharenko, SymbioForge
//
//  nfntafnx - utility provides converting PalmOS(tm) binary font resources
//  either to NFNT or AFNX data formats.
//
//  This program is free software; you can redistribute it and/or modify 
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
//  gcc nfntafnx.c toollib.c -o nfntafnx
//
//   Where GCC is a GNU C compiler or other used by your computer operation
//  system;

#include <stdio.h>
#include "toollib.h"

static		void usage( void );

main( int argc, char *argv[] )
{
  short		i, arglen, nextArgLen;
  UInt16	baseResource=1000;
  Boolean	returnStatus=0;

  puts("nfntafnx - Convert data format of PalmOS font resources between NFNT and AFNX.");
  puts("Version 1.3 (c) 2005, SymbioForge (http://www.symbioforge.com/feii/).");

  // Collect command line arguments.
  for ( i=1; i < argc; i++ )
   { 
    arglen = strlen( argv[i] ); // Get length of this argument;
    nextArgLen = (i+1) < argc ? strlen( argv[i+1] ) : 0;

    // Treat command line options;
    if ( (arglen >= 1 ) && (argv[i][0]=='-' || argv[i][0]=='/') ) // Assume this is a flag;
       {  unsigned int value=0;
          if ( arglen >= 2 ) // Check the flag;
	    {
	     switch( argv[i][1] )
               {
		case 'h':
		case 'H':
		case '?':
			usage();
		break;

		case 'O': // Already treated.
                case 'o':
			if (arglen <=2) 
			   {
                   		puts( "\nError: name of output file is absent. Program stopped.\00" );
				errorExit( errorExitValue );
			   };
                break;

                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
               };
          }
	  else
	  {
		puts( "\nError: incorrect use of argument. Program stopped.\00" );
		errorExit( errorExitValue );
	  };
      }
     else
      { FILE		* inFilePtr, * tmpFilePtr, * outFilePtr;
	char		outFileNameStr[ dosFileNameLength ];
	char 		* outFileNamePtr=NULL; // File name of output file.
	unsigned long	fileSize, dataToWriteSize=0;
	UInt16		wordValue=0, tablesLength=0, i0, i1;
	Boolean		headerFormatV5=false;
	UInt8		fontResType=0, oppositeFontResType=0;
	FontType	fontHeaderV1;
	FontTypeV2	fontHeaderV2;
	FontDensityType fontDensity[ maxDensityNumber ];

	// Get option - name of output file name.
	if ( (nextArgLen>=3) && (argv[i+1][0]=='-' || argv[i+1][0]=='/') ) // Assume this is a flag;
	   {
	     if ( (argv[i+1][1]=='o' || argv[i+1][1]=='O') ) 
		outFileNamePtr=&argv[i+1][2];
	     else
		{
                   puts( "\nError: name of output file is absent. Program stopped.\00" );
		   errorExit( errorExitValue );
		};
	   };

	tmpFilePtr = tmpfile();
	if ( tmpFilePtr==NULL )
	   {
		printf("\nError: can not create temporary file. Program aborted.\n\00");
		errorExit( errorFileOperation ); 
	   };
        fileSeek( tmpFilePtr, 0l ); // Seek file at the start position.

	inFilePtr = openFile( argv[i], "rb" );
	if ( inFilePtr==NULL )
		{ 
		  printf("\nError: can not open file '%s'. Program aborted.\n\00", argv[i] );
		  errorExit( errorFileOperation );
                };

	printf("\nConvert to "/*'%s' font to ", argv[i]*/ );

	fontResType = getFontResourceType( inFilePtr, true );
	(unsigned long)fileSize = (unsigned long)fileLength( inFilePtr ); // Get size of opened file;

	switch( fontResType )
	 {
	  case type1NFNT:
		oppositeFontResType=type1AFNX;
		getFontHeaderV1( inFilePtr, &fontHeaderV1 );
	 	tablesLength=( fontHeaderV1.lastGlyph-fontHeaderV1.firstGlyph )+3;
		putFontHeaderV1BE( &fontHeaderV1, tmpFilePtr );
	  break;

	  case type2nfnt:
		oppositeFontResType=type2afnx;
		getFontHeaderV1( inFilePtr, (FontType *)&fontHeaderV2 );
		fontHeaderV2.version=fileGetWord( inFilePtr );
		fontHeaderV2.densityCount=fileGetWord( inFilePtr );

	 	tablesLength=( fontHeaderV2.lastGlyph-fontHeaderV2.firstGlyph )+3;
		headerFormatV5=false;

		if ( fontHeaderV2.owTLoc==(((fontTag2V5Length+fontHeaderV2.densityCount*fontDensityV5Length+tablesLength*2)/2)-8) )
		     {
			headerFormatV5=true;
			putFontHeaderV2osV5BE( &fontHeaderV2, tmpFilePtr );
			for( i1=0; i1<fontHeaderV2.densityCount; i1++ )
			 {
			   getFontDensityPartV5( inFilePtr, &fontDensity[ i1 ] );
			   fontHeaderV2.reserved = 0;
		    	   putFontDensityPartV5BE( &fontDensity[ i1 ], tmpFilePtr );
		         };
		     }
		else
		     {
			// Get reserved word from the input file.
			fontHeaderV2.reserved=fileGetWord( inFilePtr ); 

			// Write afnx v2 font header data.
			putFontHeaderV2BE( &fontHeaderV2, tmpFilePtr );
			for( i1=0; i1<fontHeaderV2.densityCount; i1++ )
			 {
			   getFontDensityPart( inFilePtr, &fontDensity[ i1 ] );
		    	   putFontDensityPartBE( &fontDensity[ i1 ], tmpFilePtr );
		         };
		     };
	  break;

	  case type1AFNX:
		oppositeFontResType=type1NFNT;
		getFontHeaderV1BE( inFilePtr, &fontHeaderV1 );
	 	tablesLength=( fontHeaderV1.lastGlyph-fontHeaderV1.firstGlyph )+3;
		putFontHeaderV1( &fontHeaderV1, tmpFilePtr );
	  break;

	  case type2afnx:
		oppositeFontResType = type2nfnt;

	 	getFontHeaderV1BE( inFilePtr, (FontType *)&fontHeaderV2 );
		fontHeaderV2.version = fileGetWordBE( inFilePtr );
		fontHeaderV2.densityCount = fileGetWordBE( inFilePtr );

	 	tablesLength=( fontHeaderV2.lastGlyph-fontHeaderV2.firstGlyph )+3;
		headerFormatV5=false;

		if ( fontHeaderV2.owTLoc==(fontTag2V5Length+fontHeaderV2.densityCount*fontDensityV5Length+tablesLength*2)/2-8 )
		     {
			headerFormatV5=true;
			putFontHeaderV2osV5( &fontHeaderV2, tmpFilePtr );
			for( i1=0; i1<fontHeaderV2.densityCount; i1++ )
			 {
			   getFontDensityPartV5BE( inFilePtr, &fontDensity[ i1 ] );
			   fontHeaderV2.reserved = 0;
		    	   putFontDensityPartV5( &fontDensity[ i1 ], tmpFilePtr );
		         };
		     }
		else
		     {
			// Get reserved word from the input file.
			fontHeaderV2.reserved=fileGetWordBE( inFilePtr ); 

			// Write afnx v2 font header data.
			putFontHeaderV2( &fontHeaderV2, tmpFilePtr );
			for( i1=0; i1<fontHeaderV2.densityCount; i1++ )
			 {
			   getFontDensityPartBE( inFilePtr, &fontDensity[ i1 ] );
		    	   putFontDensityPart( &fontDensity[ i1 ], tmpFilePtr );
		         };
		     };
	  break;

	  default:
		printf("\nError: file '%s' is not a binary font resource. Program stopped.", argv[i] );
		errorExit( errorFileOperation );
	  break;
	 };

	if ( outFileNamePtr==NULL )
	 { UInt16 resourceID;
	   char * hexValueStrPtr=NULL;

	   // Attempt to retreive resource number from input name of file.
	   if ( strlen( argv[i] ) > 4 )
	    {
 	      hexValueStrPtr=&argv[i][4];
	      if ( strlen(hexValueStrPtr)!=0 )
			resourceID=(UInt16)HtoI( hexValueStrPtr );
	    }
	   else
	      resourceID=baseResource;
	   baseResource++;

	   fillAreaByByte( outFileNameStr, 0, dosFileNameLength );
	   switch( oppositeFontResType )
		{
		 case type1AFNX:
			sprintf( outFileNameStr, "%s%04x.bin", "AFNX", resourceID );
			// printf("'%s', %s.", outFileNameStr, "AFNX version 1" );
		 break;

		 case type1NFNT:
			sprintf( outFileNameStr, "%s%04x.bin", "NFNT", resourceID );
			// printf("'%s', %s.", outFileNameStr, "NFNT version 1" );
		 break;

	  	 case type2afnx:
			sprintf( outFileNameStr, "%s%04x.bin", "afnx", resourceID );
			// printf("'%s', %s, %s format.", outFileNameStr, "afnx version 2", headerFormatV5?"PalmOS(tm) v5.x":"Cobalt(tm)" );
		 break;

	  	 case type2nfnt:
			sprintf( outFileNameStr, "%s%04x.bin", "nfnt", resourceID );
			// printf("'%s', %s %s format.", outFileNameStr, "nfnt version 2", headerFormatV5?"PalmOS(tm) v5.x":"Cobalt(tm)" );
		 break;
		};
	    outFileNamePtr=outFileNameStr;
	 };

	 switch( oppositeFontResType )
	      {
		 case type1AFNX:
			printf("'%s', %s.", outFileNamePtr, "AFNX version 1" );
		 break;

		 case type1NFNT:
			printf("'%s', %s.", outFileNamePtr, "NFNT version 1" );
		 break;

	  	 case type2afnx:
			printf("'%s', %s, %s format.", outFileNamePtr, "afnx version 2", headerFormatV5?"PalmOS(tm) v5.x":"Cobalt(tm)" );
		 break;

	  	 case type2nfnt:
			printf("'%s', %s %s format.", outFileNamePtr, "nfnt version 2", headerFormatV5?"PalmOS(tm) v5.x":"Cobalt(tm)" );
		 break;
	       };


	 for( i1=0; i1 < (tablesLength /* *2 */); i1++ )
	    {
	 	if ( fontResType==type2nfnt || fontResType==type1NFNT )
		   {
			wordValue=fileGetWord(inFilePtr);
		 	filePutWordBE( tmpFilePtr, wordValue );
		   }
		else
		   {
			wordValue=fileGetWordBE(inFilePtr);
		 	filePutWord( tmpFilePtr, wordValue );
		   };
	    };

 	 // For afnx <-> nfnt convertions.
	 for( i1=0; i1 < (tablesLength /* *2 */); i1++ )
	    {
		wordValue=fileGetWord( inFilePtr );
		filePutWord( tmpFilePtr, wordValue );
	    };

	dataToWriteSize=fileSize-ftell(inFilePtr);
	proceedBinaryDataFileTransfer(inFilePtr, tmpFilePtr, (UInt32)ftell( inFilePtr), (UInt32)dataToWriteSize);
	fclose( inFilePtr ); // close input intel x86 raw font data file;

	outFilePtr=openFile( outFileNamePtr, "wb+");
	if ( outFilePtr==NULL )
	   {
		printf("\nError: can not open file to write '%s'. Program aborted.", outFileNamePtr );
		errorExit( errorFileOperation );
	   };

	// Write all data to output file.
	proceedBinaryDataFileTransfer(tmpFilePtr, outFilePtr, 0l, fileSize);
	fclose( outFilePtr );
	fclose( tmpFilePtr );
	//printf(" Done.");
       }; // End of file option;
   }; // End of command line arguments counter;

 if ( argc==1 ) 
	usage(); // Display help list;

 return returnStatus; 
};

// Display user's help.
static void usage( void )
{
  puts("\nUsage ([] means optional):");
  puts("   nfntafnx <fontres.bin> [flags]...");
  puts("               ... fontres.bin is a Palm OS(tm) binary font resource.\n");
  puts("Flags are:\n");
  puts("-o<filename>  to specify alternate name of output binary font file.");
  puts("-h, -?	      help screen.");
};
