//  Copyright (C) 2008 Alexander S. Ovcharenko, SymbioForge
//
//  pctofnt - utility to convert PC screen psf or raw font file to PalmOS(tm)
//  binary font resource type 1.
//
//  This program is free software; you can redistribute it and/or modify 
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------
//
//  To compile this file use the following command line options:
//  gcc pctofnt.c toollib.c -o pctofnt
//
//   Where GCC is a GNU C compiler or other used by your computer operation
//  system;

#include <stdio.h>
#include "toollib.h"

// Define length of font in bytes for the specified dimensions.

static void usage( void );

main( int argc, char *argv[] )
{ int		returnStatus = 0;
  int   	i, i1, i2, i3, arglen;
  int		noFileName = 1;
  unsigned int  resBeginFrom = 1000;
  UInt8		glyphImageByte;

  puts("pctofnt - PC screen font data to PalmOS(tm) v1 font resource.");
  puts("Version 1.4 (c) 2008, SymbioForge (http://www.symbioforge.com/feii/).");

  for ( i=1; i < argc; i++ )
   { 
    arglen = strlen( argv[i] ); // Get length of this argument;

    // Treat command line options;
    if ( (arglen >= 1 ) && (argv[i][0]=='-' || argv[i][0]=='/') ) // Assume this is a flag;
       {  unsigned int value=0;
          if ( arglen >= 2 ) // Check the flag;
	    {
	     switch( argv[i][1] )
               {
		case 'h':
		case 'H':
		case '?':
			usage();
		break;

		case 'N':
                case 'n':
			if ( arglen >=3 ) // Get value;
			 {
			  value=atoi( &argv[i][2] );
			  if ( value < 1000 )
				{
                             	 puts( "\nError: resource value is less than 1000. Program stopped.\00" );
				 errorExit( errorExitValue );
				}
			  else
				{
				 resBeginFrom = value;
				};
		         }
                        else
			 {
                          puts( "\nError: resource start value is not set. Program stopped.\00" );
			  errorExit( errorExitValue );
			 };
                break;

                default:
			puts( "\nError: unrecognized flag used. Program aborted.\00" );
			errorExit( errorExitValue );
                break;
               };
          };
      }
     else
      { FILE		* inFilePtr, * tmpFilePtr, * outFilePtr;
	char		* resFileNamePtr=NULL;
	unsigned long	fileSize;
	UInt16		tablesLength, columnBytes;
	FontType	fontV1Header;
	psfFontV1Type	psf1FontHeader;
	psfFontV2Type	psf2FontHeader;
	UInt8		fileFormat=0; // Where 0 - raw data, 1-PSF version 1, 2-PSF version -2.

	inFilePtr = openFile( argv[i], "rb" );
	if ( inFilePtr == NULL )
		{ 
		  printf("\nError: can not open file '%s'. Program aborted.\n\00", argv[i] );
		  errorExit( errorFileOperation );
                };

	(unsigned long)fileSize = (unsigned long)fileLength( inFilePtr ); // Get size of opened file;

	// Check for PSF font format.
	psf2FontHeader.magic[0]=psf1FontHeader.magic[0]=fileGetByteWFName( inFilePtr, argv[ i ] );
	psf2FontHeader.magic[1]=psf1FontHeader.magic[1]=fileGetByteWFName( inFilePtr, argv[ i ] );
	if ( psf1FontHeader.magic[0]==PSF1_MAGIC0 && psf1FontHeader.magic[1]==PSF1_MAGIC1 )
		fileFormat=1;
	else
	   {
		psf2FontHeader.magic[2]=fileGetByteWFName( inFilePtr, argv[ i ] );
		psf2FontHeader.magic[3]=fileGetByteWFName( inFilePtr, argv[ i ] );

		if ( psf2FontHeader.magic[0]==PSF2_MAGIC0 &&  psf2FontHeader.magic[1]==PSF2_MAGIC1 
			&& psf2FontHeader.magic[2]==PSF2_MAGIC2 && psf2FontHeader.magic[3]==PSF2_MAGIC3 )
			  fileFormat=2;
	   };

	// lets read font header in case magic bytes are for PSF font format.
	switch ( fileFormat )
	 {
	  case 1:
		psf1FontHeader.mode = fileGetByteWFName( inFilePtr, argv[i] );		
		psf1FontHeader.charsize = fileGetByteWFName( inFilePtr, argv[i] ); // Font Height.		

		// Check file length to be accordant to the calculated size.
		if ((sizeof(psfFontV1Type)+(psf1FontHeader.charsize*256)) > fileSize )
		  fileFormat=0;
		else
		  fontV1Header.fRectHeight=psf1FontHeader.charsize;

		// puts("\nFont version 1.");
		// printf("\nFont mode: %u", psf1FontHeader.mode );
		// printf("\nChar size: %u\n", psf1FontHeader.charsize );
		// fclose( inFilePtr ); // close input intel x86 raw font data file;
		// errorExit( errorFileOperation );
	  break;
	
	  case 2: 
		printf("\nError: PSF font version 2 is not supported.\n\00");
		fclose( inFilePtr ); // close input intel x86 raw font data file;
		errorExit( errorFileOperation );
	  break;

	 };

	// Determine font dimension by file size.
	if ( fileFormat==0 )
	  {
		switch( fileSize )
	 		{
	  		 case fontLength08x08:
				fontV1Header.fRectHeight=8;
	  		 break;

	  		 case fontLength08x14:
				fontV1Header.fRectHeight=14;
	  		 break;

	  		 case fontLength08x16:
				fontV1Header.fRectHeight=16;
	  		 break;

	  		 case fontLength08x18:
				fontV1Header.fRectHeight=18;
	  		 break;

	  		 case fontLength08x19:
				fontV1Header.fRectHeight=19;
	  		 break;

	  		 default:
				printf("\nError: unknown length of PC screen font file. Program stopped.\n\00");
				fclose( inFilePtr ); // close input intel x86 raw font data file;
				errorExit( errorFileOperation );
	  		break;
		       };
	   };

	// Set header fixed values.
        fontV1Header.fontType   				= 0x9000;
        fontV1Header.fRectWidth=fontV1Header.maxWidth	= 8;
        fontV1Header.firstGlyph   				= dosFirstByte;
        fontV1Header.lastGlyph   				= dosLastByte;
        fontV1Header.nDescent				= 0;
        fontV1Header.descent					= 2;				// Set descent for the font header is allways 2.
        fontV1Header.ascent					= fontV1Header.fRectHeight-2;       // ascent is a height minus two.
        fontV1Header.leading					= 0;
        fontV1Header.kernMax				= 0;

	// Calculate dynamic header values;
	tablesLength = (fontV1Header.lastGlyph-fontV1Header.firstGlyph)+3;
	fontV1Header.rowWords				= ((UInt32)((tablesLength-1)*(UInt32)fontV1Header.fRectWidth)+15L)/16L;
	fontV1Header.owTLoc				= (UInt32)fontV1Header.rowWords*(UInt32)fontV1Header.fRectHeight+(UInt32)tablesLength+5L;

	tmpFilePtr = tmpfile();
	if ( tmpFilePtr==NULL )
	   {
          	 printf("\nError: can not open temporary file. Program aborted.\n\00");
	  	 errorExit( errorFileOperation );
	   };

     	 // Store font v1 header to temporary file.
	fileSeek( tmpFilePtr, 0l ); // Seek file at the start position.
	putFontHeaderV1( (FontType *)&fontV1Header, tmpFilePtr );
	
	// Write glyph images.
	for( i1=0; i1< fontV1Header.fRectHeight; i1++ )
	 {
	  columnBytes=0;
	  for( i2=0; i2 < tablesLength-1; i2++ )
	    {
		if ( i2 < (tablesLength-2) )
		  {
			// Seek file at the glyph image part.
	       		fileSeek( inFilePtr, (i2*fontV1Header.fRectHeight+i1)+(fileFormat==0)?0:sizeof(psfFontV1Type) ); 
			glyphImageByte=fileGetByteWFName( inFilePtr, argv[ i ] );
        		// fileSeek( tmpFilePtr, i2+(UInt32)i1*(UInt32)fontV1Header.rowWords*2l ); 
			filePutByte( tmpFilePtr, glyphImageByte );
			columnBytes++;
		  }
		else
		  {
			// Add defauld byte image;
			if ( (i1==1) || (i1==fontV1Header.fRectHeight-2) )
			   filePutByte( tmpFilePtr, 0x7e );
			else
			if ( (i1>1) && (i1< fontV1Header.fRectHeight-2) )
			   filePutByte( tmpFilePtr, 0x42 );
			else
			   filePutByte( tmpFilePtr, 0x0 );
			columnBytes++;

			// Write padding bytes;
			for ( ; columnBytes<(fontV1Header.rowWords*2); columnBytes++ )
				filePutByte( tmpFilePtr, 0xFF ); 
		  }
	    };
	  };

	
	// Add tables to file.
	for( i1=0; i1 < tablesLength; i1++ )
		filePutWord( tmpFilePtr, (i1 * fontV1Header.fRectWidth) );
	for( i1=0; i1 < tablesLength-1; i1++ )
		filePutWord( tmpFilePtr, fontV1Header.fRectWidth );
	filePutWord( tmpFilePtr, 0 ); // Padding 0 byte for the width offset table.
	(unsigned long)fileSize = (unsigned long)fileLength( tmpFilePtr );

	if ( (psf1FontHeader.mode & PSF1_MODE512) !=0 )
		puts("\nWARNING: Only 256 characters of 512 are exported to Palm OS font.");
	if ( (psf1FontHeader.mode & PSF1_MODEHASTAB) !=0 )
      		puts("\nWARNING: Unicode information can not be saved to Palm OS font.");

     	// Move file from temporary file to binary font resource;
	resFileNamePtr=(char *)malloc(dbNameLength+1);
	resFileNamePtr[0]=0;

	outFilePtr = openFile( resFileNamePtr, "wb+" );
	if ( outFilePtr == NULL )
	   {
	    printf("\nError: can not write to file '%s'. Program aborted.\n\00", "out file name" );
	    errorExit( errorFileOperation );
	   }
	else
	   {
	    printf("\nConvert '%s' to '%s'...", argv[ i ], resFileNamePtr );
	    proceedBinaryDataFileTransfer( tmpFilePtr, outFilePtr, 0l, (unsigned long)fileSize );
	    printf(" Done.");
	   };
	fclose( outFilePtr );
	free( resFileNamePtr );

	fclose( tmpFilePtr );
	fclose( inFilePtr ); // close input intel x86 raw font data file;
	resBeginFrom++; // Increase resource number.
      }; // End of if "file" condition;
 }; // End of command line arguments counter;

//puts("");

if ( argc==1 ) 
	usage(); // Display help list;

 return returnStatus; 
};

// Display user's help.
static void usage( void )
{
  puts("\nUsage ([] means optional):");
  puts("    pctofnt [flags] <pcfont>...");
  puts("               ... where pcfont is an PSF v1 or x86 raw font data files.\n");
  puts("                   Determines font dimensions from the lengths of input files:");
  puts("                   8x8 - 2048 bytes; 8x14 - 3584 bytes; 8x16 - 4096 bytes;");
  puts("                   8x19 - 4864 bytes.\n");
  puts("Flags are:\n");
  puts("-n<value>  <value> is a resource number for output file name.");
  puts("           Resource filename number is 1000 (NFNT03e8.bin) by default.\n");
  puts("-h, -?	   help screen.");
};
