//  Copyright (C) 2004 Alexander S. Ovcharenko, SymbioForge.
//
//  toollib.c - is a library for Handy Pilot Font Editor's font utility.
//  It is essential part at compilation time.
//
//    This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//    This program is a part of Handy Pilot Font Editor II (Raster font
//  editor for PalmOS powered handheld computers). Sources may be distributed
//  separately with no lost of original idea and functionality under the
//  terms of the GNU General Public License.
// 
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------

#include <stdio.h>
#include <time.h>
#include "toollib.h"
// Close all opened files and exit with errorCode;
void errorExit( UInt16 errorCode )
{
 fcloseall(); // Close all opened files;
 exit( errorCode );
};

// Opens file with file name for the specified mode;
FILE * openFile(const char * fileNameStrPtr, char * fileOpenModeStrPtr )
{ FILE * fileHandle;

  if ( fileNameStrPtr==NULL )
    {
	printf("\nFile name is not set. Program stopped.\n\00", fileNameStrPtr );
   	errorExit( errorExitValue );
    }
  else
  if ((fileHandle = fopen( fileNameStrPtr, fileOpenModeStrPtr ))!=NULL)
  	return fileHandle;
  else
    {
	printf("\nError: Can not open file '%s'. Program aborted.\n\00", fileNameStrPtr );
	errorExit( errorFileOperation );
    };
};

// Seek file (filePtr) position at (filePos).
void fileSeek( FILE * filePtr, unsigned long filePos )
{ int errorCode;
  errorCode = fseek( filePtr, filePos, SEEK_SET );
  if ( errorCode < 0 || (unsigned long)ftell( filePtr )!= (unsigned long)filePos )
	{
	  puts("\nError: Unable to seek file. Program aborted.\00");
	  errorExit( errorFileOperation );
	};
};

// Retreive the size of opened file;
unsigned long fileLength( FILE * filePtr )
{ unsigned long oldFilePos, fileLength;
  int errorCode;

  (unsigned long)oldFilePos = (unsigned long)ftell( filePtr );
  errorCode = fseek( filePtr, 0, SEEK_END ); // Seek to 0 offset from the end of file.
  if ( errorCode < 0 )
	{
	  puts("\nError: Can not get the length of file. Program aborted.\00");
	  errorExit( errorFileOperation );
	}

  (unsigned long)fileLength=(unsigned long)ftell( filePtr ); // Get position of end of file;
  fileSeek( filePtr, (unsigned long)oldFilePos );

  return fileLength;
};

// Read one byte from file; File position will changes on one character plus;
UInt8 fileGetByte( FILE * filePtr )
{ int byteValue;

 byteValue = fgetc( filePtr );
 if ( byteValue==-1 )
    {
      puts("\nError: Can not read from file. Program aborted.\00");  
      errorExit( errorFileOperation );
    };
 return (UInt8)byteValue;
};

// Read byte from file. Say about file name when error occur.
// Added to be more informative.
UInt8 fileGetByteWFName( FILE * filePtr, char * fileNameStrPtr )
{ int byteValue;

 if (fileNameStrPtr==NULL)
    {
      printf("\nError: reading operation failed - invalid file name. Program aborted.\n\00");  
      errorExit( errorFileOperation );
    };

 byteValue = fgetc( filePtr );
 if ( byteValue==-1 )
    {
      printf("\nError: Can not read from '%s' file (off: %LX). Program aborted.\n\00", fileNameStrPtr, ftell( filePtr ) );  
      errorExit( errorFileOperation );
    };
 return (UInt8)byteValue;
};

// Write one byte from file; File position will changes on one character plus;
void filePutByte( FILE * filePtr, UInt8 byteToWrite )
{
 if ( fputc( byteToWrite, filePtr )==-1 )
  {
    puts("\nError: Can not write to file. Program aborted.\00");
    errorExit( errorFileOperation );
  };
  return;
};

// Added to be more informative.
UInt16 fileGetWordWFName( FILE * filePtr, char * fileNameStrPtr )
{ UInt16 wordValue;
  wordValue = fileGetByteWFName( filePtr, fileNameStrPtr )*0x100;
  wordValue+= fileGetByteWFName( filePtr, fileNameStrPtr );
  return wordValue;
};

// Read one word from file; File position will changes on two characters plus;
UInt16 fileGetWord( FILE * filePtr )
{ UInt16 wordValue;
  wordValue  = fileGetByte( filePtr )*0x100;
  wordValue += fileGetByte( filePtr );
  return wordValue;
};

// Read big endian word from file.
UInt16 fileGetWordBE( FILE * filePtr )
{ UInt16 wordValue;

  wordValue = fileGetByte( filePtr );
  wordValue+= fileGetByte( filePtr )*0x100;

  return wordValue;
};

// Write one word to file; File position will changes on two characters plus;
void filePutWord( FILE * filePtr, UInt16 wordToWrite )
{
 filePutByte( filePtr, wordToWrite/0x100 );
 filePutByte( filePtr, wordToWrite%0x100 );
 return;
};

// Put big endian word to file.
void filePutWordBE( FILE * filePtr, UInt16 wordToWrite )
{
 filePutByte( filePtr, wordToWrite%0x100 );
 filePutByte( filePtr, wordToWrite/0x100 );
 return;
};

// Read double word from file; File position will increased on four characters;
UInt32 fileGetDWord( FILE * filePtr )
{ UInt32 dwordValue;
  (UInt32)dwordValue = (UInt32)0x10000*(UInt32)fileGetWord( filePtr );
  (UInt32)dwordValue += (UInt32)fileGetWord( filePtr );
  return (UInt32)dwordValue;
};

// Read big endian double word from file.
UInt32 fileGetDWordBE( FILE * filePtr )
{ UInt32 dwordValue;
  (UInt32)dwordValue = (UInt32)fileGetWordBE( filePtr );
  (UInt32)dwordValue += (UInt32)0x10000 * (UInt32)fileGetWordBE( filePtr );
  return (UInt32)dwordValue;
};

// Write DWord value to the opened file.
void filePutDWord( FILE * filePtr, UInt32 dwordToWrite )
{
  filePutWord( filePtr, (UInt16)((UInt32)dwordToWrite/(UInt32)0x10000) );
  filePutWord( filePtr, (UInt16)((UInt32)dwordToWrite%(UInt32)0x10000) );
  return;
};

// Write DWord value to the opened file.
UInt32 filePutDWordBE( FILE * filePtr, UInt32 dwordToWrite )
{
  filePutWordBE( filePtr, (UInt16)((UInt32)dwordToWrite%(UInt32)0x10000) );
  filePutWordBE( filePtr, (UInt16)((UInt32)dwordToWrite/(UInt32)0x10000) );
  return;
};

// Read .pdb Palm OS(tm) database header from file;
void fileGetPdbHeader( pdbHeaderType * pdbHeaderPtr, FILE * filePtr)
{ unsigned long oldFilePos;

  (unsigned long)oldFilePos = (unsigned long)ftell( filePtr );
  fileSeek( filePtr, 0l ); // Go to the begining of the opened file;;

  if ( fread( &(pdbHeaderPtr->dbName), 1, dbNameLength, filePtr ) != dbNameLength )
	{
      	 puts("\nError: Can not read from file. Program aborted.\00");
	 errorExit( errorFileOperation );
	};
  pdbHeaderPtr->dbName[ dbNameLength-1 ] = 0; // Terminate string by NULL
                                                                        
  pdbHeaderPtr->flags = fileGetWord( filePtr );
  pdbHeaderPtr->version = fileGetWord( filePtr );

  (UInt32)pdbHeaderPtr->createTime = (UInt32)fileGetDWord( filePtr );
  (UInt32)pdbHeaderPtr->modifiedTime = (UInt32)fileGetDWord( filePtr );
  (UInt32)pdbHeaderPtr->backupTime = (UInt32)fileGetDWord( filePtr );

  (UInt32)pdbHeaderPtr->modNumber = (UInt32)fileGetDWord( filePtr );
  (UInt32)pdbHeaderPtr->appInfo = (UInt32)fileGetDWord( filePtr );
  (UInt32)pdbHeaderPtr->sortInfo = (UInt32)fileGetDWord( filePtr );

  (UInt32)pdbHeaderPtr->dbType = (UInt32)fileGetDWord( filePtr );
  (UInt32)pdbHeaderPtr->id = (UInt32)fileGetDWord( filePtr );
  (UInt32)pdbHeaderPtr->idSeed = (UInt32)fileGetDWord( filePtr );

  // Get records list.
  pdbHeaderPtr->recordList.nextRecordList = fileGetDWord( filePtr );
  pdbHeaderPtr->recordList.numRecords = fileGetWord( filePtr );
  pdbHeaderPtr->recordList.firstEntry = fileGetWord( filePtr );

  fileSeek( filePtr, (unsigned long)oldFilePos );
};

// Write Palm OS Database header data to file.
void filePutPdbHeader( FILE * filePtr, pdbHeaderType * pdbHeaderPtr )
{ unsigned long oldFilePos;
  (unsigned long)oldFilePos = (unsigned long)ftell( filePtr );

  // Write database name
  proceedBinaryDataFileOut( filePtr, (void *)&pdbHeaderPtr->dbName, 0l, dbNameLength );

  fileSeek( filePtr, dbNameLength ); // Go to the begining of the opened file;;
  filePutWord( filePtr, (UInt16)pdbHeaderPtr->flags ); // Store database flags;
  filePutWord( filePtr, (UInt16)pdbHeaderPtr->version ); // Store database version;
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->createTime ); // Store database creation time;
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->modifiedTime ); // 
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->backupTime );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->modNumber );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->appInfo );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->sortInfo );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->dbType );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->id );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->idSeed );
  filePutDWord( filePtr, (UInt32)pdbHeaderPtr->recordList.nextRecordList );
  filePutWord( filePtr, (UInt16)pdbHeaderPtr->recordList.numRecords );
  // filePutWord( filePtr, (UInt16)pdbHeaderPtr->recordList.firstEntry );

  fileSeek( filePtr, (unsigned long)oldFilePos );
};

// Return offset of pdbRecordListType entry by number of entry defined in the listIndex.
UInt32 fileGetPDBListEntryOffset( UInt16 listIndex, FILE * filePtr )
{ unsigned long		oldFilePos, filePos;
  pdbRecordListType 	listEntry; 

  (UInt32)oldFilePos = (UInt32)ftell( filePtr );
  // Get first (0) list entry in the .pdb;
  (UInt32)filePos = ( (UInt32)lengthOfpdbHeader - (UInt32)lengthOfpdbRecordList );

  // Get linst entry offset in .pdb file by index. Leaved for further improvements.
  //(UInt32)filePos += (UInt32)lengthOfpdbRecordList * listIndex;

  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;
  return (UInt32)filePos;
};

// Get offset to next List entry;
UInt32 fileGetPDBNextRecordListOffset( UInt32 currentListRecordOffset, FILE * filePtr ) 
{ UInt32 nextRecordListOffset=0, oldFilePos;

(UInt32)oldFilePos = (UInt32)ftell( filePtr );
fileSeek( filePtr, (UInt32)currentListRecordOffset ); // Return to previous file position;
(UInt32)nextRecordListOffset = (UInt32)fileGetDWord( filePtr );
fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;

return (UInt32)nextRecordListOffset;
};

// return number of records for specified recordList;
UInt16 fileGetPDBNumRecords( UInt16 listIndex, FILE * filePtr )
{ UInt32	oldFilePos, listIndexOffset;
  UInt16	recordsNumber=0;

  (UInt32)oldFilePos = (UInt32)ftell( filePtr );
  (UInt32)listIndexOffset = (UInt32)fileGetPDBListEntryOffset( listIndex, filePtr );

  fileSeek( filePtr, listIndexOffset ); // Return to previous file position;
  (UInt32)listIndexOffset=(UInt32)fileGetDWord( filePtr );
  (UInt16)recordsNumber = (UInt16)fileGetWord( filePtr );
  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;

  return recordsNumber;
};

// Find format of data type.
// Returns:
// -1 if data type have unknown format.
// 0 if data type are resources
// 1 if data are records.
int fileGetPDBResourceFormat( UInt32 listEntryOffset, FILE * filePtr )
 { UInt32		oldFilePos=0l, filePos, nextListEntryOffset;
   UInt32		resourcesDataTopBound=0, recordsDataTopBound=0, dataBottomBound=0;
   UInt16		dataPartsNumber;
   pdbRecordListType	listEntry;
   int dataFormat = unknownDataFormat;

   (UInt32)oldFilePos = (UInt32)ftell( filePtr );
   dataBottomBound=fileLength( filePtr );

   // Get top bounds for different format.
   fileSeek( filePtr, (UInt32)listEntryOffset ); // Seek file at next list entry position;
   (UInt32)nextListEntryOffset=(UInt32)fileGetDWord( filePtr );
   dataPartsNumber = fileGetWord( filePtr );

   if ( dataPartsNumber > 0 ) // No records/resources in this database.
      { unsigned int i;
	Boolean	 dataCheckPass;

   	(UInt32)recordsDataTopBound= (UInt32)listEntryOffset+lengthOfpdbRecordList-2l+( (UInt32)lengthOfpdbRecordEntry * dataPartsNumber );
	(UInt32)resourcesDataTopBound=(UInt32)listEntryOffset+lengthOfpdbRecordList-2l+( (UInt32)lengthOfpdbResourceEntry * dataPartsNumber );

//	printf("Top bound of records data: %lx\n", (UInt32)recordsDataTopBound );
//	printf("Top bound of resources data: %lx\n", (UInt32)resourcesDataTopBound );
//	printf("Bottom bound of data: %lx\n", dataBottomBound );
	
	dataCheckPass = true;
	// Exam records offsets to data bound.
	for( i=0; i < dataPartsNumber; i++ )
	 {
	  filePos= (UInt32)listEntryOffset+(UInt32)lengthOfpdbRecordList-2l+( (UInt32)lengthOfpdbRecordEntry * (UInt32)i );
	  if ( filePos < dataBottomBound-sizeof(UInt32) )
	     {
	  	fileSeek( filePtr, (UInt32)filePos ); // Seek to record data;
		(UInt32)filePos = (UInt32)fileGetDWord( filePtr ); // Get offset of record;
		if ( (UInt32)filePos >= ((UInt32)recordsDataTopBound) && ( (UInt32)filePos < (UInt32)dataBottomBound) ) 
			continue;
		dataCheckPass=false;
		break;
	     }
	  else
	     {
		dataCheckPass=false;
		break;
	     };
	 };

	if ( dataCheckPass )
		  dataFormat = recordsDataFormat;

	dataCheckPass = true;
	// Exam resources offsets to data bounds.
	for( i=0; i < dataPartsNumber; i++ )
	 {
	  // Get offset for the next resource data.
	  (UInt32)filePos= (UInt32)listEntryOffset+(UInt32)lengthOfpdbRecordList-2l+( (UInt32)lengthOfpdbResourceEntry * (UInt32)i )+(UInt32)lengthOfpdbResourceEntry-4l;

	 // printf("Offset of resource: %u is %lx\n", i, filePos );
	  if ( filePos < dataBottomBound-sizeof(UInt32) )
	     {
	  	fileSeek( filePtr, (UInt32)filePos ); // Seek to record data;
	  	(UInt32)filePos = (UInt32)fileGetDWord( filePtr ); // Get offset of record;
	  	if ( filePos>=((UInt32)resourcesDataTopBound) && (filePos<dataBottomBound) ) 
			continue;
	  	dataCheckPass=false;
	  	break;
	     }
	   else
	     {
		dataCheckPass=false;
		break;
	     };
	};
	if ( dataCheckPass )
		dataFormat = resourcesDataFormat;

      }; // End of ( dataParts > 0 );

   fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;
   return	dataFormat;
 };

// Obtain an offset of a record (recordIndex) in an opened file.
// Returns offset of requested record number in a file, 0 if none.
UInt32 fileGetPDBRecordOffset( UInt32 listEntryOffset, UInt16 recordIndex, FILE * filePtr )
{ UInt32		oldFilePos, filePos=0l, fileSize=0l, totalIndexCounter=0l;
  unsigned		int recordRealIndex=0l, counter, fontType;
  pdbRecordListType 	listEntry; 
  pdbRecordEntryType	recordEntry;
  pdbHeaderType		pdbHeader;

  (UInt32)oldFilePos = (UInt32)ftell( filePtr );

  fileSeek( filePtr, (UInt32)listEntryOffset ); // Seek file at next list entry position;
  (UInt32)listEntry.nextRecordList = (UInt32)fileGetDWord( filePtr );
  listEntry.numRecords = fileGetWord( filePtr );
  (UInt32)recordEntry.recordOffset = 0l;
  if ( recordIndex < listEntry.numRecords ) 
    {
        (UInt32)filePos = (UInt32)listEntryOffset+lengthOfpdbRecordList-2+( (UInt32)lengthOfpdbRecordEntry * recordIndex );
   	fileSeek( filePtr, (UInt32)filePos ); // Seek file at desired record entry position;
  	(UInt32)recordEntry.recordOffset = (UInt32)fileGetDWord( filePtr ); // Get offset of record;
    };
  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;
  return ((UInt32)recordEntry.recordOffset);
};

// Returns:
// + Offset of requested resource number in a file, 0 if none.
// + Resource type in typePtr if it is not a NULL;
// + Resource id in idPtr if it is not a NULL;
UInt32 fileGetResourceOffset( UInt32 listEntryOffset, UInt16 resourceIndex, FILE * filePtr, UInt32 * typePtr, UInt16 * idPtr )
{ UInt32		oldFilePos, fileSize=0l, filePos, totalIndexCounter=0l;
  unsigned		int recordRealIndex=0l, counter, fontType;
  pdbRecordListType 	listEntry; 
  resourceEntryType	resourceEntry;
  pdbHeaderType		pdbHeader;

  (UInt32)oldFilePos = (UInt32)ftell( filePtr );

  fileSeek( filePtr, (UInt32)listEntryOffset ); // Seek file at next list entry position;
  (UInt32)listEntry.nextRecordList = (UInt32)fileGetDWord( filePtr );
  listEntry.numRecords = fileGetWord( filePtr );
  (UInt32)resourceEntry.offset = (UInt32)resourceEntry.type = 0l;
  resourceEntry.id = 0;

  if ( resourceIndex < listEntry.numRecords ) 
    {
      (UInt32)filePos = (UInt32)listEntryOffset+lengthOfpdbRecordList-2+( (UInt32)resourceEntryLength * resourceIndex );
      fileSeek( filePtr, (UInt32)filePos ); // Seek file at desired record entry position;

      (UInt32)resourceEntry.type = (UInt32)fileGetDWord( filePtr ); // Get offset of record;
      (UInt16)resourceEntry.id = (UInt16)fileGetWord( filePtr ); // Get offset of record;
      (UInt32)resourceEntry.offset = (UInt32)fileGetDWord( filePtr ); // Get offset of record;
    };
  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;

  if ( typePtr!=NULL ) (UInt32)*typePtr=(UInt32)resourceEntry.type;
  if ( idPtr!=NULL ) (UInt16)*idPtr=(UInt16)resourceEntry.id;

  return ((UInt32)resourceEntry.offset);
};

// Proceed data write 
void proceedBinaryDataFileOut( FILE * outFilePtr, void * dataPtr, unsigned long filePosition, unsigned long dataLength )
{ unsigned long oldFilePos;

if (dataLength==0l) return;

(unsigned long)oldFilePos = (unsigned long)ftell( outFilePtr );
fileSeek( outFilePtr, (unsigned long)filePosition ); // Seek to read position;
if ( fwrite( dataPtr, 1, (unsigned long)dataLength, outFilePtr )!=dataLength )
	{
    	  puts("\nError: Can not write to file. Program aborted.\00");
  	  errorExit( errorFileOperation );
	};
fileSeek( outFilePtr, (unsigned long)oldFilePos ); // Return to previous file position;

return;
};

// Reads data from  offset and length from inFilePtr. Writes to file - outFilePtr;
void proceedBinaryDataFileTransfer( FILE * inFilePtr, FILE * outFilePtr, UInt32 fileDataOffset, UInt32 dataSize )
{ unsigned long	oldFilePos, filePos;
  unsigned int	bytesToRead;
  char		fileBufferPtr[ bufferLengthInBytes ];

  (unsigned long)oldFilePos = (unsigned long)ftell( inFilePtr );

  for( (unsigned long)filePos=0l; (unsigned long)filePos < (unsigned long)dataSize; (unsigned long)filePos += bufferLengthInBytes )
   {
	bytesToRead = ((unsigned long)filePos+(unsigned long)bufferLengthInBytes) < (unsigned long) dataSize ? bufferLengthInBytes:(unsigned long)dataSize-(unsigned long)filePos;
  	fileSeek( inFilePtr, (unsigned long)filePos+(unsigned long)fileDataOffset ); // Seek to read position;
  	if ( fread( fileBufferPtr, 1, bytesToRead, inFilePtr )==bytesToRead )
      	   { 
		// Proceed buffer write;
		// fileSeek( outFilePtr, (unsigned long)filePos ); // Seek to write position;
		if ( fwrite( fileBufferPtr, 1, bytesToRead, outFilePtr )!=bytesToRead )
		  {
	  	    // free( fileBufferPtr );
	  	    puts("\nError: Can not write to file. Program aborted.\00");
      	  	    errorExit( errorFileOperation );
		  };
      	   }
  	else
      	   {
	  	// free( fileBufferPtr );
	  	puts("\nError: Can not read record. Program aborted.\00");
      	  	errorExit( errorFileOperation );
      	   };
  };
  // free( fileBufferPtr );
  fileSeek( inFilePtr, (unsigned long)oldFilePos ); // Return to previous file position;
};

// Save font header. Assume filePtr is an opened file for write mode.
Boolean putFontHeaderV1( FontType * fontHeaderPtr, FILE * filePtr )
{
 if ( fontHeaderPtr==NULL ) return false; // Do not save header when no buffers.
 filePutWord( filePtr, fontHeaderPtr->fontType );
 filePutWord( filePtr, fontHeaderPtr->firstGlyph );
 filePutWord( filePtr, fontHeaderPtr->lastGlyph );
 filePutWord( filePtr, fontHeaderPtr->maxWidth );
 filePutWord( filePtr, fontHeaderPtr->kernMax );
 filePutWord( filePtr, fontHeaderPtr->nDescent );
 filePutWord( filePtr, fontHeaderPtr->fRectWidth );
 filePutWord( filePtr, fontHeaderPtr->fRectHeight );
 filePutWord( filePtr, fontHeaderPtr->owTLoc );
 filePutWord( filePtr, fontHeaderPtr->ascent );
 filePutWord( filePtr, fontHeaderPtr->descent );
 filePutWord( filePtr, fontHeaderPtr->leading );
 filePutWord( filePtr, fontHeaderPtr->rowWords );

 return true; // Say true if can not keep silence.
};

// Save font header. Assume filePtr is an opened file for write mode.
Boolean putFontHeaderV1BE( FontType * fontHeaderPtr, FILE * filePtr )
{
 if ( fontHeaderPtr==NULL ) return false; // Do not save header when no buffers.
 filePutWordBE( filePtr, fontHeaderPtr->fontType );
 filePutWordBE( filePtr, fontHeaderPtr->firstGlyph );
 filePutWordBE( filePtr, fontHeaderPtr->lastGlyph );
 filePutWordBE( filePtr, fontHeaderPtr->maxWidth );
 filePutWordBE( filePtr, fontHeaderPtr->kernMax );
 filePutWordBE( filePtr, fontHeaderPtr->nDescent );
 filePutWordBE( filePtr, fontHeaderPtr->fRectWidth );
 filePutWordBE( filePtr, fontHeaderPtr->fRectHeight );
 filePutWordBE( filePtr, fontHeaderPtr->owTLoc );
 filePutWordBE( filePtr, fontHeaderPtr->ascent );
 filePutWordBE( filePtr, fontHeaderPtr->descent );
 filePutWordBE( filePtr, fontHeaderPtr->leading );
 filePutWordBE( filePtr, fontHeaderPtr->rowWords );
 return true; // Say true if can not keep silence.
};

// Write Font Header type version 2 of PalmOS 5.x (Original fontType + 2 words).
Boolean putFontHeaderV2osV5( FontTypeV2 * fontHeaderV2Ptr, FILE * filePtr )
{
 putFontHeaderV1( (FontType *)fontHeaderV2Ptr, filePtr );
 filePutWord( filePtr, fontHeaderV2Ptr->version );
 filePutWord( filePtr, fontHeaderV2Ptr->densityCount );
 return true;
};

// Write Font Header type version 2 to file. (Original fontType + 2 addition word + reserved ).
Boolean putFontHeaderV2( FontTypeV2 * fontHeaderV2Ptr, FILE * filePtr )
{
 putFontHeaderV1( (FontType *)fontHeaderV2Ptr, filePtr );
 filePutWord( filePtr, fontHeaderV2Ptr->version );
 filePutWord( filePtr, fontHeaderV2Ptr->densityCount );
 filePutWord( filePtr, fontHeaderV2Ptr->reserved );
 return true;
};

// Write Font Header type version 2 to file. (Original fontType + 2 words of additional fileds).
Boolean putFontHeaderV2osV5BE( FontTypeV2 * fontHeaderV2Ptr, FILE * filePtr )
{
 putFontHeaderV1BE( (FontType *)fontHeaderV2Ptr, filePtr );
 filePutWordBE( filePtr, fontHeaderV2Ptr->version );
 filePutWordBE( filePtr, fontHeaderV2Ptr->densityCount );
 return true;
};

// Write Font Header type version 2 to file. (Original fontType + 3 words of additional fileds).
Boolean putFontHeaderV2BE( FontTypeV2 * fontHeaderV2Ptr, FILE * filePtr )
{
 putFontHeaderV1BE( (FontType *)fontHeaderV2Ptr, filePtr );
 filePutWordBE( filePtr, fontHeaderV2Ptr->version );
 filePutWordBE( filePtr, fontHeaderV2Ptr->densityCount );
 filePutWordBE( filePtr, fontHeaderV2Ptr->reserved );
 return true;
};


// Write Density record to the file.
Boolean putFontDensityPart( FontDensityType * fontDensityPtr, FILE * filePtr )
{
 filePutWord( filePtr, fontDensityPtr->density );
 filePutWord( filePtr, fontDensityPtr->reserved );
 filePutDWord( filePtr, fontDensityPtr->glyphBitsOffset );
 return true;
};

// Write Density record to the file.
Boolean putFontDensityPartV5( FontDensityType * fontDensityPtr, FILE * filePtr )
{
 filePutWord( filePtr, fontDensityPtr->density );
 filePutDWord( filePtr, fontDensityPtr->glyphBitsOffset );
 return true;
};


Boolean putFontDensityPartBE( FontDensityType * fontDensityPtr, FILE * filePtr )
{
 filePutWordBE( filePtr, fontDensityPtr->density );
 filePutWordBE( filePtr, fontDensityPtr->reserved );
 filePutDWordBE( filePtr, fontDensityPtr->glyphBitsOffset );
 return true;
};

// Write density part v5 Big Endian.
Boolean putFontDensityPartV5BE( FontDensityType * fontDensityPtr, FILE * filePtr )
{
 filePutWordBE( filePtr, fontDensityPtr->density );
 filePutDWordBE( filePtr, fontDensityPtr->glyphBitsOffset );
 return true;
};


// Get density parts from the file.
Boolean getFontDensityPartV5( FILE * filePtr, FontDensityType * fontDensityPtr )
{
 fontDensityPtr->density=fileGetWord(filePtr);
 fontDensityPtr->reserved=0; // Skip reserved word when old PalmOS v5.x format used.
 (UInt32)fontDensityPtr->glyphBitsOffset=(UInt32)fileGetDWord(filePtr);
};

Boolean getFontDensityPart( FILE * filePtr, FontDensityType * fontDensityPtr )
{
 fontDensityPtr->density=fileGetWord(filePtr);
 fontDensityPtr->reserved=fileGetWord(filePtr); 
 (UInt32)fontDensityPtr->glyphBitsOffset=(UInt32)fileGetDWord(filePtr);
};

// Get density parts from the file.
Boolean getFontDensityPartV5BE( FILE * filePtr, FontDensityType * fontDensityPtr )
{
 fontDensityPtr->density=fileGetWordBE( filePtr );
 // fontDensityPtr->reserved=fileGetWordBE( filePtr );
 (UInt32)fontDensityPtr->glyphBitsOffset=(UInt32)fileGetDWordBE(filePtr);
};

// Get density parts from the file.
Boolean getFontDensityPartBE( FILE * filePtr, FontDensityType * fontDensityPtr )
{
 fontDensityPtr->density=fileGetWordBE( filePtr );
 fontDensityPtr->reserved=fileGetWordBE( filePtr );
 (UInt32)fontDensityPtr->glyphBitsOffset=(UInt32)fileGetDWordBE(filePtr);
};

// Load font header. Assume filePtr is a binary font record type 1.
Boolean getFontHeaderV1( FILE * filePtr, FontType * fontHeaderPtr )
{ 
if ( fontHeaderPtr==NULL ) return false; // Do not load header when no buffer.
fontHeaderPtr->fontType=fileGetWord( filePtr );
fontHeaderPtr->firstGlyph=fileGetWord( filePtr );
fontHeaderPtr->lastGlyph=fileGetWord( filePtr );
fontHeaderPtr->maxWidth=fileGetWord( filePtr );
fontHeaderPtr->kernMax=fileGetWord( filePtr );
fontHeaderPtr->nDescent=fileGetWord( filePtr );
fontHeaderPtr->fRectWidth=fileGetWord( filePtr );
fontHeaderPtr->fRectHeight=fileGetWord( filePtr );
fontHeaderPtr->owTLoc=fileGetWord( filePtr );
fontHeaderPtr->ascent=fileGetWord( filePtr );
fontHeaderPtr->descent=fileGetWord( filePtr );
fontHeaderPtr->leading=fileGetWord( filePtr );
fontHeaderPtr->rowWords=fileGetWord( filePtr );

return true;
};

// Load afnx font header. Assume filePtr is a binary font record type 1.
Boolean getFontHeaderV1BE( FILE * filePtr, FontType * fontHeaderPtr )
{ 
if ( fontHeaderPtr==NULL ) return false; // Do not load header when no buffer.

fontHeaderPtr->fontType=fileGetWordBE( filePtr );
fontHeaderPtr->firstGlyph=fileGetWordBE( filePtr );
fontHeaderPtr->lastGlyph=fileGetWordBE( filePtr );
fontHeaderPtr->maxWidth=fileGetWordBE( filePtr );
fontHeaderPtr->kernMax=fileGetWordBE( filePtr );
fontHeaderPtr->nDescent=fileGetWordBE( filePtr );
fontHeaderPtr->fRectWidth=fileGetWordBE( filePtr );
fontHeaderPtr->fRectHeight=fileGetWordBE( filePtr );
fontHeaderPtr->owTLoc=fileGetWordBE( filePtr );
fontHeaderPtr->ascent=fileGetWordBE( filePtr );
fontHeaderPtr->descent=fileGetWordBE( filePtr );
fontHeaderPtr->leading=fileGetWordBE( filePtr );
fontHeaderPtr->rowWords=fileGetWordBE( filePtr );

return true;
};


// Load font header. Says about file name when error occur.
Boolean getFontHeaderV1WFName( FILE * filePtr, FontType * fontHeaderPtr, char * fileNameStrPtr )
{
 if ( fontHeaderPtr==NULL ) return false; // Do not load header when no buffer.
 fontHeaderPtr->fontType=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->firstGlyph=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->lastGlyph=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->maxWidth=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->kernMax=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->nDescent=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->fRectWidth=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->fRectHeight=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->owTLoc=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->ascent=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->descent=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->leading=fileGetWordWFName( filePtr, fileNameStrPtr );
 fontHeaderPtr->rowWords=fileGetWordWFName( filePtr, fileNameStrPtr );

 return true;
};

// Converts font version to it's resource type.
UInt32 fontVersionToResType( UInt8 fontVersion )
{ UInt32 resourceType=0;
  switch( fontVersion )
	{
	 case type1NFNT:
	  	(UInt32)resourceType=(UInt32)fontVersion1;
	 break;
	 case type2nfnt:
	  	(UInt32)resourceType=(UInt32)fontVersion2;
	 break;
	 case type1AFNX:
	  	(UInt32)resourceType=(UInt32)fontVersion3;
	 break;
	 case type2afnx:
	  	(UInt32)resourceType=(UInt32)fontVersion4;
	 break;
	};
 return (UInt32)resourceType;
};


// Returns type of font resource:
// 1 - type version 1, 2 - type version 2.
// 3 - for type AFNX version 1, 4 -afnx type version 2;
// 0 - either resource is not a font, or is not a resource.
UInt8 getFontResourceType( FILE * filePtr, Boolean doDetailedCheck )
{ unsigned long	oldFilePos;
  UInt8		returnValue = 0;
  UInt16	chkFontVersion = 0, fontType=0;
  FontType	* fontV1HeaderPtr = NULL;
  FontTypeV2	fontV2Header;
  Boolean	isAFNX=false;

  (unsigned long)oldFilePos = (unsigned long)ftell( filePtr );

  fileSeek( filePtr, 0l ); // Go at the begining of opened file;

  fontV1HeaderPtr = (FontType *)&fontV2Header; // Get address of FontType ver.2 data.
  fontV1HeaderPtr->fontType = fontType = fileGetWord( filePtr );
  isAFNX = (fontType==(UInt16)0x0090) || (fontType==(UInt16)0x0092);

  if ( (fontType==(UInt16)0x9000) || (fontType==(UInt16)0x9200) || isAFNX )
     {
	fileSeek( filePtr, 0l ); // Go at the begining of opened file;
	if ( isAFNX )
		getFontHeaderV1BE( filePtr, fontV1HeaderPtr );
	else
		getFontHeaderV1( filePtr, fontV1HeaderPtr ); // Get AFNX font header data.
	returnValue = isAFNX ? type1AFNX : type1NFNT;

	if ( fontV1HeaderPtr->fontType == (Int16)0x9200 ) // Read V2 addition data.
	  {
		fontV2Header.version = isAFNX ? fileGetWordBE( filePtr ) : fileGetWord( filePtr );
		if ( fontV2Header.version !=1 && doDetailedCheck )
		   {
  	    	     puts("\nError: font resource version is incorrect. Program aborted.\00");
      	    	     errorExit( errorExitValue );
		   };
		fontV2Header.densityCount = isAFNX ? fileGetWordBE( filePtr ) : fileGetWord( filePtr );
		returnValue = isAFNX ? type2afnx : type2nfnt;
	  };

	// Check for some font header values.
	if ( (fontV1HeaderPtr->firstGlyph > 255) ||
	     (fontV1HeaderPtr->lastGlyph > 255) ||
             (fontV1HeaderPtr->lastGlyph < fontV1HeaderPtr->firstGlyph) )
	   {
  	     puts("\nError: invalid font header values. Program stopped.\00");
  	     errorExit( errorExitValue );
	   };
	if ( fontV1HeaderPtr->fRectHeight == 0 || fontV1HeaderPtr->fRectHeight >255 )
	   {
  	     puts("\nError: invalid glyph height value. Program stopped.\00");
  	     errorExit( errorExitValue );
	   };
     };
  fileSeek( filePtr, (unsigned long)oldFilePos ); // Return to previous file position;
  return returnValue;
};

// Get offset of resource entry.
UInt32 computateResourceEntryOffset( UInt16 resourceIndex )
{ 
 return (UInt32)( (UInt32)pdbHeaderLength-(UInt32)2+(UInt32)(resourceIndex*resourceEntryLength) ); 
};

// Get resource entry data for the specified number of resource;
void fileGetResourceEntry( FILE * filePtr, resourceEntryType * resourceEntryPtr, UInt16 resourceIndex )
{ UInt32	oldFilePos, filePos;
  (UInt32)oldFilePos = (UInt32)ftell( filePtr );
  (UInt32)filePos = (UInt32)computateResourceEntryOffset( resourceIndex );
  fileSeek( filePtr, (UInt32)filePos ); // Return to previous file position;
  (UInt32)resourceEntryPtr->type = (UInt32)fileGetDWord( filePtr );
  (UInt16)resourceEntryPtr->id = (UInt16)fileGetWord( filePtr );
  (UInt32)resourceEntryPtr->offset = (UInt32)fileGetDWord( filePtr );
  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;
  // printf("Resource id: %u, resource offset %lu\n", resourceEntryPtr->id, resourceEntryPtr->offset );
  return;
};

// Store resource entry data at the desired location;
void filePutResourceEntry( resourceEntryType * resourceEntryPtr, UInt16 resourceIndex, FILE * filePtr )
{ UInt32	oldFilePos, filePos;
  (UInt32)oldFilePos = (UInt32)ftell( filePtr );

  // (UInt32)filePos = ( (UInt32)lengthOfpdbHeader-(UInt32)2 )+(UInt32)( (UInt32)resourceIndex * (UInt32)resourceEntryLength );
  (UInt32)filePos = (UInt32)computateResourceEntryOffset( resourceIndex );
  fileSeek( filePtr, (UInt32)filePos );

  filePutDWord( filePtr, (UInt32)resourceEntryPtr->type );
  filePutWord( filePtr, (UInt16)resourceEntryPtr->id );
  filePutDWord( filePtr, (UInt32)resourceEntryPtr->offset );
  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;
  return;
};

// Store record entry data. Computate by index;
void filePutRecordEntry( pdbRecordEntryType * recordEntryPtr, UInt16 recordIndex, FILE * filePtr )
{ UInt32	oldFilePos, filePos;
  (UInt32)oldFilePos = (UInt32)ftell( filePtr );

  (UInt32)filePos=((UInt32)lengthOfpdbHeader-(UInt32)2 )+(recordIndex * lengthOfpdbRecordEntry);
  fileSeek( filePtr, (UInt32)filePos );

  filePutDWord( filePtr, (UInt32)recordEntryPtr->recordOffset );
  filePutByte( filePtr, (UInt16)recordEntryPtr->attributes );
  filePutByte( filePtr, (UInt32)recordEntryPtr->uniqueID[0] );
  filePutByte( filePtr, (UInt32)recordEntryPtr->uniqueID[1] );
  filePutByte( filePtr, (UInt32)recordEntryPtr->uniqueID[2] );

  fileSeek( filePtr, (UInt32)oldFilePos ); // Return to previous file position;
  return;
};

void fillAreaByByte( void * areaPtr, unsigned char fillByte, unsigned long lenght )
{ UInt32 counter;
  if ( areaPtr==NULL ) return;

  #ifdef __ZTC__
     for( counter=0; counter < lenght; counter++ ) ((unsigned char *)areaPtr)[counter]=fillByte;
  #else
     memset( areaPtr, fillByte, lenght );
  #endif
};

// Converts hexadecimal string to decimal value;
UInt32 HtoI( char * inputStringPtr )
{ UInt32 decimalValue;
  char	 hexStr[]="0123456789ABCDEF";
  short	 stringLength, stringCounter, hexStrCounter;

  stringLength = strlen( inputStringPtr );
  stringLength = stringLength > 8 ? 8 : stringLength;
  decimalValue = 0;  

  if ( stringLength > 0 )
     {
  	for( stringCounter=0; stringCounter < stringLength; stringCounter++ )
    	  {
	   for( hexStrCounter=0; ( hexStrCounter < strlen(hexStr) ) && (tolower(inputStringPtr[stringCounter]) != tolower(hexStr[hexStrCounter]) ); hexStrCounter++ ) {};
	   if ( hexStrCounter >= strlen(hexStr) ) break; // Exit if no hexadecimal character found;
	     else
           (UInt32)decimalValue = ( (UInt32)decimalValue * 0x10 ) + hexStrCounter;
	  };
    };
 return (UInt32)decimalValue;
};

UInt32 timePalm( void )
{
    UInt32	timeForPalmOS=0l;
    time_t	sysTime;

    time( &sysTime );
    return (time_t)sysTime!=(time_t)-1 ? (UInt32)sysTime+(UInt32)0x7c25b080: (UInt32)-1;
};

// Allocate memory for glyph.
glyphBufferType * glyphBufferCreate( UInt16 bufWidth, UInt16 bufHeight )
{ glyphBufferType * glyphBufferPtr=NULL;
  size_t	bufferLengthValue;
  UInt16	widthInBytes;

  widthInBytes   = (bufWidth+7)>>3;	// Convert bits to bytes;
  bufferLengthValue = sizeof( glyphBufferType )+( widthInBytes * bufHeight );
  glyphBufferPtr = (glyphBufferType *)malloc( bufferLengthValue );
  memset( glyphBufferPtr, 0, bufferLengthValue );

  glyphBufferPtr->widthInBytes = widthInBytes;
  glyphBufferPtr->width = bufWidth;
  glyphBufferPtr->height = bufHeight;
  glyphBufferPtr->bufferPtr = (UInt8 *)( glyphBufferPtr+1 );

  return glyphBufferPtr;
};

// Scalling source glyph to buffer glyph;
// Source took from Bitmap Scaller Trainer by Tumblin / Bodies In Motion 
// Brezenhaim alike algorithm.
void glyphBufferScale( glyphBufferType * destinationGlyphBufferPtr, glyphBufferType * sourceGlyphBufferPtr )
{
UInt32 xPosSum = 0, yPosSum = 0;
UInt16 xPosSrc = 0, yPosSrc = 0;
UInt16 xPosDst = 0, yPosDst = 0;
UInt8  * glyphDataPtr = NULL;

if ( sourceGlyphBufferPtr==NULL || destinationGlyphBufferPtr==NULL ) return;

// Make glyph fatter and taller;
// It does two operations with glyph only.
if ( (destinationGlyphBufferPtr->width > sourceGlyphBufferPtr->width) && (destinationGlyphBufferPtr->height > sourceGlyphBufferPtr->height) )
 {
	for( yPosSum=yPosDst=0; yPosDst < destinationGlyphBufferPtr->height; yPosDst++ )
	  {
	   yPosSum += sourceGlyphBufferPtr->height;
	   if( yPosSum > destinationGlyphBufferPtr->height )
		  {
			yPosSum -= destinationGlyphBufferPtr->height;
			yPosSrc++;
		  };

	   for( xPosSrc=xPosSum=xPosDst = 0; xPosDst < destinationGlyphBufferPtr->width; xPosDst++ )
		  {
			xPosSum += sourceGlyphBufferPtr->width;
			if( xPosSum > destinationGlyphBufferPtr->width )
			   {
				xPosSum -= destinationGlyphBufferPtr->width;
				xPosSrc++;
			   };
			 // Fix values if they are greater of bounds;
			 if ( xPosSrc > sourceGlyphBufferPtr->width-1 ) xPosSrc = sourceGlyphBufferPtr->width-1; 
			 if ( yPosSrc > sourceGlyphBufferPtr->height-1 ) yPosSrc = sourceGlyphBufferPtr->height-1;

			 glyphDataPtr = sourceGlyphBufferPtr->bufferPtr + ( yPosSrc * sourceGlyphBufferPtr->widthInBytes) + (xPosSrc>>3); 
			 if ( (*glyphDataPtr & (0x80>>(xPosSrc & 7))) !=0 )
			   {
	  			 glyphDataPtr = destinationGlyphBufferPtr->bufferPtr + (yPosDst * destinationGlyphBufferPtr->widthInBytes) + (xPosDst>>3); 
	  			 *glyphDataPtr |= ( 1 << ( 7 - ( xPosDst&0x7 ) ) );
			   };
		   };
	  };
 }
else 
// skinnier and shorter destination bitmap;
if ( (destinationGlyphBufferPtr->width <= sourceGlyphBufferPtr->width) && (destinationGlyphBufferPtr->height <= sourceGlyphBufferPtr->height) )
  {
	for( yPosSum=yPosDst=yPosSrc=0; yPosSrc < sourceGlyphBufferPtr->height; yPosSrc++ )
	  {
	   yPosSum += destinationGlyphBufferPtr->height;
	   if( yPosSum > sourceGlyphBufferPtr->height )
		  {
			yPosSum -= sourceGlyphBufferPtr->height;
			yPosDst++;
		  };


	    for( xPosSrc=xPosSum=xPosDst=0; xPosSrc < sourceGlyphBufferPtr->width; xPosSrc++ )
		  {
			xPosSum += destinationGlyphBufferPtr->width;
			if( xPosSum > sourceGlyphBufferPtr->width )
			   {
				xPosSum -= sourceGlyphBufferPtr->width;
			    	xPosDst++;
			   };
			 // Fix values if they are greater of bounds; Nup-Nuphers if happened;
			 if ( xPosDst > destinationGlyphBufferPtr->width-1 ) xPosDst = destinationGlyphBufferPtr->width-1; 
			 if ( yPosDst > destinationGlyphBufferPtr->height-1 ) yPosDst = destinationGlyphBufferPtr->height-1;

			 glyphDataPtr = sourceGlyphBufferPtr->bufferPtr + ( yPosSrc * sourceGlyphBufferPtr->widthInBytes) + (xPosSrc>>3); 
			 if ( (*glyphDataPtr & (0x80>>(xPosSrc & 7))) !=0 )
			   {
	  			 glyphDataPtr = destinationGlyphBufferPtr->bufferPtr + (yPosDst * destinationGlyphBufferPtr->widthInBytes) + (xPosDst>>3); 
	  			 *glyphDataPtr |= ( 1 << ( 7 - ( xPosDst&0x7 ) ) );
			   };
		   };
  	 };
  };
return;
};

// Returns scaled value by percent.
UInt16 scaleValueByPercent( UInt16 value, UInt16 percent, Boolean useCeiling )
{
 UInt32 scaledValue=0;

 while(false) if ( percent==100 ) return value;

 (UInt32)scaledValue=(UInt32)((UInt32)value * (UInt32)percent);

 if ((UInt32)scaledValue!=0 )
   {
    if ( ( (useCeiling) && (percent>=100) ) || ( (!useCeiling) && (percent < 100) ) )
      (UInt32)scaledValue=(UInt32)scaledValue/100l;
    else
      (UInt32)scaledValue=((UInt32)scaledValue/100l)+( ((UInt32)scaledValue%100l)!=0l ? 1l : 0l );
    }

 return (UInt16)((UInt32)scaledValue > (UInt32)0xFFFF ? -1 : (scaledValue &0xFFFF));
};

// Returns scaled previous value by percent.
UInt16 scaleBackValueByPercent( UInt16 value, UInt16 percent, Boolean useCeiling )
{
 UInt32 scaledValue=0;

 while(false) if ( percent==100 ) return value;

 (UInt32)scaledValue=(UInt32)((UInt32)value * 100l);

 if ((UInt32)scaledValue!=0l )
   {
    if ( ( (useCeiling)&&(percent>=100) ) || ((!useCeiling)&&(percent<100)) )
      (UInt32)scaledValue=((UInt32)scaledValue/percent)+( ((UInt32)scaledValue%percent)!=0l ? 1l : 0l );
    else
      (UInt32)scaledValue=(UInt32)scaledValue/percent;
   };

 return (UInt16)((UInt32)scaledValue > (UInt32)0xFFFF ? -1 : ((UInt32)scaledValue &0xFFFF));
};
