//  Copyright (C) 2004 Alexander S. Ovcharenko, SymbioForge
//  toollib.h - is a header file of toollib.c library.
//
//   This program is free software; you can redistribute it and/or modify it
//  under the terms of the GNU General Public License as published by the
//  FREE SOFTWARE FUNDATION; either version 2 of the License or (at your
//  opinion) any later version.
//
//  To obtain a complete GNU License from internet use URL http://www.fsf.org
//  -------------------------------------------------------------------------

#define useCeilingWhenScaleWidth	true
#define useCeilingWhenScaleHeight	true

// Definition of some constants:
#define		fntExtendedFormatMask	0x200

#ifndef		TOOL_LIBRARY

#ifndef		true
 #define	true			1
#endif
#ifndef		false
 #define	false			0
#endif

#ifndef		max
 #define		max(a,b)	( ( (a) > (b) ) ? (a) : (b) )
#endif
#ifndef		min
 #define		min(a,b)	( ( (a) < (b) ) ? (a) : (b) )
#endif

// Definitions for the PC Screen fixed font sizes.
#define		dosFirstByte				0
#define		dosLastByte				255

#define		fontLength08x08				2048
#define		fontLength08x14				3584
#define		fontLength08x16				4096
#define		fontLength08x18				4608
#define		fontLength08x19				4864


// Font types.
#define		typeUnknown				0
#define		type1NFNT				1
#define		type2nfnt				2
#define		type1AFNX				3
#define		type2afnx				4

#define		maxDensityNumber			4

// Density values for FontTypeV2 font parts.
#define		singleDensityValue			72
#define		oneAndAHalfDensityValue			108
#define		doubleDensityValue			144
#define		tripleDensityValue			216
#define		quadrupleDensityValue			288

#define		singleDensityPercent			100
#define		oneAndAHalfDensityPercent		150
#define		doubleDensityPercent			200
#define		tripleDensityPercent			300
#define		quadrupleDensityPercent			400

#define		oneAndAHalfToSingleDensityPercent	 75
#define		doubleToSingleDensityPercent		 50
#define		tripleToSingleDensityPercent		 33
#define		quadrupleToSingleDensityPercent		 25

#define		TOOL_LIBRARY		1
#define		dbNameLength		32

#define		filePathLength		128

#define		errorExitValue		255
#define		errorFileOperation	254
#define		bufferLengthInBytes	1024

#define		unknownDataFormat	-1
#define		recordsDataFormat	1
#define		resourcesDataFormat	2
#define		dosFileNameLength	8+1+3+1 // 12 bytes file name + /0.

typedef		signed char		Int8;
typedef		signed short		Int16;
typedef		signed long		Int32;

typedef		unsigned char		Boolean;
typedef 	unsigned char		UInt8;
typedef		unsigned short		UInt16;
typedef		unsigned long		UInt32;

typedef struct
	{
	  UInt32		nextRecordList;    // A position to next list of records;
	  UInt16		numRecords;	// Number of records in the list;
	  UInt16		firstEntry; 	// An array of record entries ( pdbRecordEntryType );
	} pdbRecordListType;
#define 	lengthOfpdbRecordList	8

typedef struct
	{
	  UInt32	recordOffset;    // offset to the record from the begining of file.
	  UInt8		attributes;      // record attributes;
	  UInt8		uniqueID[ 3 ];   // uniqueID (3 bytes) long.
	} pdbRecordEntryType;		 // 
#define		lengthOfpdbRecordEntry	8

typedef struct
	{
	 UInt32		type;		// A type of resource;
	 UInt16		id;		// Id of resource;
	 UInt32		offset;		// Offset to resource.
	} resourceEntryType;

#define		resourceEntryLength		10
#define		lengthOfpdbResourceEntry	resourceEntryLength

// A header of .pdb files;
typedef struct 
	{
	 char 			dbName[ dbNameLength ];	// Name of a .pdb database; (32 bytes)
	 UInt16			flags;		// Flags of a database; (2 bytes)
	 UInt16			version;	// Database version. (2 bytes)
	 UInt32			createTime;	// The time in seconds database created; (4 bytes) 40
	 UInt32			modifiedTime;	// Last the time database modified; (4 bytes)
	 UInt32			backupTime;	// Time for backup. (4 bytes)
	 UInt32			modNumber;	// Number of modification; (4 bytes)

	 UInt32			appInfo;	// Application info; (4 bytes)
	 UInt32			sortInfo;	// Sorting information; (4 bytes) 60
	 UInt32			dbType;		// Type of Database 'appl' for executables; (4 bytes); 
	 UInt32			id;		// 'Appl' or DB creator ID; (4 bytes);
	 UInt32			idSeed;		// Unique ID seed; (4 bytes);
	 pdbRecordListType	recordList;	// First RecordList entry; ( 8 bytes );
	} pdbHeaderType;

#define		pdbHeaderLength		80
#define		lengthOfpdbHeader	pdbHeaderLength

typedef	struct FontCharInfoTag
	{
	  Int8 		offset;
	  Int8		width;
	} FontCharInfoType;

typedef struct
	{
	 UInt16		renamedCategories;
	 char		categoryLabels[16][16];
	 char		categoryUniqIDs[16];
	 UInt8		lastUniqID;
	 UInt8		padding;
	} pdbApplicationInfoType;

// A modified copy of font header definitions from <PalmOS SDK>Font.h
typedef struct FontTagV1
{
	Int16 fontType; 		// font type
	Int16 firstGlyph; 		// ASCII code of first glyph
	Int16 lastGlyph; 		// ASCII code of last glyph
	Int16 maxWidth; 		// maximum character width
	Int16 kernMax; 			// negative of maximum character kern
	Int16 nDescent; 		// negative of descent
	Int16 fRectWidth; 		// width of font rectangle
	Int16 fRectHeight; 		// height of font rectangle
	Int16 owTLoc; 			// offset to offset/width table
	Int16 ascent; 			// ascent
	Int16 descent; 			// descent
	Int16 leading; 			// leading
	Int16 rowWords; 		// row width of bit image / 2
} FontType;
#define fontTag1Length			26

// Font v2 density definition for PalmOS(tm) versions 5.x.
typedef struct FontDensityTagV5	// Definition of font density element of header. 
{
	Int16	density;
	UInt32	glyphBitsOffset;
} FontDensityV5Type;
#define		fontDensityV5Length	6

typedef struct FontTagV2osV5
{
	// first part is basically the same as NFNT font record
	Int16	fontType;
	Int16	firstGlyph;		  // ASCII code of first font glyph
	Int16	lastGlyph;		  // ASCII code of last font glyph
	Int16	maxWidth;		  // widMax = maximum character width
	Int16	kernMax;		  // negative of maximum character kern
	Int16	nDescent;		  // negative of descent
	Int16	fRectWidth;		  // width of font rectangle
	Int16	fRectHeight;		  // height of font rectangle
	Int16	owTLoc;			  // offset to offset/width table
	Int16	ascent;			  // ascent
	Int16	descent;		  // descent
	Int16	leading;		  // leading
	Int16	rowWords;		  // row width of bit image / 2

	// New fields (if fntExtendedFormatMask is set)
	Int16	version;		    // 1 = PalmNewFontVersion
	Int16	densityCount;		    // 
	//  FontDensityV5Type densities[0]; // array of 1 or more records
} FontTypeV2osV5;
#define		fontTag2V5Length	30


// NOTICE: reserved fields are found in Palm OS(tm) Cobalt documentation. 
// In the previous version of documentations this fileds are absent.
typedef struct FontDensityTag	// Definition of font density element of header. 
{
	Int16	density;
	Int16	reserved;
	UInt32	glyphBitsOffset;
} FontDensityType;
#define		fontDensityLength	8
#define		fontDensityV6Length	fontDensityLength


typedef struct FontTagV2
{
	// first part is basically the same as NFNT font record
	Int16	fontType;
	Int16	firstGlyph;		  // ASCII code of first font glyph
	Int16	lastGlyph;		  // ASCII code of last font glyph
	Int16	maxWidth;		  // widMax = maximum character width
	Int16	kernMax;		  // negative of maximum character kern
	Int16	nDescent;		  // negative of descent
	Int16	fRectWidth;		  // width of font rectangle
	Int16	fRectHeight;		  // height of font rectangle
	Int16	owTLoc;			  // offset to offset/width table
	Int16	ascent;			  // ascent
	Int16	descent;		  // descent
	Int16	leading;		  // leading
	Int16	rowWords;		  // row width of bit image / 2

	// New fields (if fntExtendedFormatMask is set)
	Int16	version;		  // 1 = PalmNewFontVersion
	Int16	densityCount;		  // 
	Int16	reserved;		  // (new in prg reference on cobalt OS).
	//  FontDensityType densities[0]; // array of 1 or more records
} FontTypeV2;
#define		fontTag2Length	32
#define		fontTag2V6Length	fontTag2Length


#define PSF1_MAGIC0     0x36
#define PSF1_MAGIC1     0x04

#define PSF1_MODE512    0x01
#define PSF1_MODEHASTAB 0x02
#define PSF1_MODEHASSEQ 0x04
#define PSF1_MAXMODE    0x05

#define PSF1_SEPARATOR  0xFFFF
#define PSF1_STARTSEQ   0xFFFE

#define PSF2_MAGIC0     0x72
#define PSF2_MAGIC1     0xb5
#define PSF2_MAGIC2     0x4a
#define PSF2_MAGIC3     0x86

// bits used in flags
#define PSF2_HAS_UNICODE_TABLE 0x01

// max version recognized so far
#define PSF2_MAXVERSION 0

// UTF8 separators
#define PSF2_SEPARATOR  0xFF
#define PSF2_STARTSEQ   0xFE

typedef struct 
		{
        		UInt8 magic[2];     // PSF font file prefix.
        		UInt8 mode;         // PSF font mode
        		UInt8 charsize;     // 
		} psfFontV1Type;

typedef struct 
		{
        		UInt8	magic[4];
        		UInt16	version;
        		UInt16	headersize;    /* offset of bitmaps in file */
        		UInt16	flags;
        		UInt16	length;        /* number of glyphs */
        		UInt16	charsize;      /* number of bytes for each character */
        		UInt16	height, width; /* max dimensions of glyphs */
        		// charsize = height * ((width + 7) / 8)
		} psfFontV2Type;


typedef struct // Define type for buffer of one glyph;
		{
			UInt16	widthInBytes;
			Int16	width;
			Int16	height;
			UInt8	* bufferPtr;
		} glyphBufferType;

// BufferLength is amount of bytes to use with glyph operations;
#define  rowLength		16
#define  maxGlyphColumnHeight	127
#define	 bufferLength		(maxGlyphColumnHeight * rowLength) + sizeof( glyphBufferType )+4 // pixels 127*127 + header size;
#define  maxTableElements	258

//Define font database ids. Some old compilers does not treat symbolic constants.
#define		fontDatabaseID0		0x464f4e54	// 'FONT';
#define		fontDatabaseID1		0x466f6e74	// 'Font'
#define		fontDatabaseID2		0x666f6e74	// 'font'
#define		applDatabaseID0		0x6170706c	// 'appl'

#define		fontVersion1		0x4e464e54	// 'NFNT'
#define		fontVersion2		0x6e666e74	// 'nfnt'
#define		fontVersion3		0x41464e58	// 'AFNX'
#define		fontVersion4		0x61666e78	// 'afnx'

#define		fontsForOS5		0x464f5335	// 'FOS5';

extern void errorExit( UInt16 );
extern FILE * openFile( const char *, char * );
extern void fileSeek( FILE *, UInt32 );
extern UInt32 fileLength( FILE * );
extern unsigned char fileGetByte( FILE * );
extern unsigned char fileGetByteWFname( FILE *, char * );

extern UInt16	fileGetWord( FILE * );
extern UInt16	fileGetWordWFName( FILE *, char * );
extern UInt16	fileGetWordBE( FILE * );

extern UInt32 fileGetDWord( FILE * );

extern void	filePutByte( FILE *, UInt8 );
extern void	filePutWord( FILE *, UInt16 );
extern void	filePutDWord( FILE *, UInt32 );

extern void	filePutWordBE( FILE *, UInt16 );

extern void	fileGetPdbHeader( pdbHeaderType *, FILE * );
extern void	filePutPdbHeader( FILE *, pdbHeaderType * );
extern UInt16	fileGetPDBNumRecords( UInt16, FILE * );
extern UInt32	fileGetPDBListEntryOffset( UInt16, FILE * );
extern UInt32   fileGetPDBNextRecordListOffset( UInt32, FILE * );
extern UInt32	fileGetPDBRecordOffset( UInt32, UInt16, FILE * );
extern UInt32	fileGetResourceOffset( UInt32, UInt16, FILE *, UInt32 *, UInt16 * );
extern void	filePutRecordEntry( pdbRecordEntryType *, UInt16, FILE *);
extern int	fileGetPDBResourceFormat( UInt32 listEntryOffset, FILE * filePtr );
extern void	proceedBinaryDataFileOut( FILE *, void *, unsigned long, unsigned long );
extern void	proceedBinaryDataFileTransfer( FILE *, FILE *, UInt32, UInt32);
extern Boolean	getFontHeaderV1( FILE *, FontType * );
extern Boolean	getFontHeaderV1BE( FILE *, FontType * );
extern Boolean	getFontHeaderV1WFName( FILE *, FontType *, char * );
extern Boolean	putFontHeaderV1( FontType *, FILE * );
extern Boolean	putFontHeaderV1BE( FontType *, FILE * );

extern Boolean  putFontHeaderV2osV5( FontTypeV2 *, FILE * );
extern Boolean  putFontHeaderV2osV5BE( FontTypeV2 *, FILE * );
extern Boolean  putFontDensityPartV5( FontDensityType *, FILE * );
extern Boolean  putFontDensityPartV5BE( FontDensityType *, FILE * );

extern Boolean	putFontHeaderV2( FontTypeV2 *, FILE * );
extern Boolean	putFontHeaderV2BE( FontTypeV2 *, FILE * );

extern Boolean	putFontDensityPart( FontDensityType *, FILE * );

extern Boolean  putFontDensityPartBE( FontDensityType *, FILE * );


extern Boolean	getFontDensityPart( FILE *, FontDensityType *);
extern Boolean  getFontDensityPartV5( FILE *, FontDensityType * );
extern Boolean  getFontDensityPartV5BE( FILE *, FontDensityType * );

extern Boolean	getFontDensityPartBE( FILE *, FontDensityType *);
extern UInt8	getFontResourceType( FILE *, Boolean );
extern void	fileGetResourceEntry( FILE *, resourceEntryType *, UInt16 );
extern void	filePutResourceEntry( resourceEntryType *, UInt16, FILE * );
extern void     fillAreaByByte( void *, unsigned char, unsigned long );
extern UInt32	fontVersionToResType( UInt8 );
extern UInt32	HtoI( char * );
extern UInt32	timePalm( void );

extern glyphBufferType * glyphBufferCreate( UInt16, UInt16 );
extern void	glyphBufferScale( glyphBufferType *, glyphBufferType *);
extern UInt16   scaleValueByPercent( UInt16, UInt16, Boolean );
extern UInt16   scaleBackValueByPercent( UInt16, UInt16, Boolean );

#endif // End of TOOL_LIBRARY definitions
